"""This module contains data objects meant to be serialized between
host processes and user processes. They should contain no
business logic or clever indexing. Use the classes in external.py
for that.
"""

import inspect
import json
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from enum import Enum
from typing import Any, Final, NamedTuple, Optional, Union, cast

from typing_extensions import Self, TypeAlias

from dagster import (
    StaticPartitionsDefinition,
    _check as check,
)
from dagster._config.pythonic_config import (
    ConfigurableIOManagerFactoryResourceDefinition,
    ConfigurableResourceFactoryResourceDefinition,
)
from dagster._config.pythonic_config.resource import coerce_to_resource, is_coercible_to_resource
from dagster._config.snap import ConfigFieldSnap, ConfigSchemaSnapshot, snap_from_config_type
from dagster._core.definitions import (
    AssetSelection,
    JobDefinition,
    PartitionsDefinition,
    RepositoryDefinition,
    ScheduleDefinition,
)
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_job import is_reserved_asset_job_name
from dagster._core.definitions.asset_sensor_definition import AssetSensorDefinition
from dagster._core.definitions.asset_spec import AssetExecutionType
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.automation_condition_sensor_definition import (
    AutomationConditionSensorDefinition,
)
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionSnapshot,
)
from dagster._core.definitions.definition_config_schema import ConfiguredDefinitionConfigSchema
from dagster._core.definitions.dependency import (
    GraphNode,
    Node,
    NodeHandle,
    NodeOutputHandle,
    OpNode,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.metadata import (
    MetadataFieldSerializer,
    MetadataMapping,
    MetadataValue,
    TextMetadataValue,
    normalize_metadata,
)
from dagster._core.definitions.multi_dimensional_partitions import MultiPartitionsDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.partition import DynamicPartitionsDefinition, ScheduleType
from dagster._core.definitions.partition_mapping import (
    PartitionMapping,
    get_builtin_partition_mapping_types,
)
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.resource_requirement import ResourceKeyRequirement
from dagster._core.definitions.schedule_definition import DefaultScheduleStatus
from dagster._core.definitions.sensor_definition import (
    DefaultSensorStatus,
    SensorDefinition,
    SensorType,
)
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from dagster._core.definitions.utils import DEFAULT_GROUP_NAME
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.origin import RepositoryPythonOrigin
from dagster._core.snap import JobSnap
from dagster._core.snap.mode import ResourceDefSnap, build_resource_def_snap
from dagster._core.storage.io_manager import IOManagerDefinition
from dagster._core.storage.tags import COMPUTE_KIND_TAG
from dagster._core.utils import is_valid_email
from dagster._record import IHaveNew, record, record_custom
from dagster._serdes import whitelist_for_serdes
from dagster._serdes.serdes import (
    FieldSerializer,
    get_prefix_for_a_serialized,
    is_whitelisted_for_serdes_object,
)
from dagster._time import datetime_from_timestamp
from dagster._utils.error import SerializableErrorInfo
from dagster._utils.warnings import suppress_dagster_warnings

DEFAULT_MODE_NAME = "default"
DEFAULT_PRESET_NAME = "default"

# Historically, SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE could on the metadata of an asset
# to encode the execution type of the asset.
SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE = "dagster/asset_execution_type"


@whitelist_for_serdes(
    storage_name="ExternalRepositoryData",
    storage_field_names={
        "schedules": "external_schedule_datas",
        "partition_sets": "external_partition_set_datas",
        "sensors": "external_sensor_datas",
        "asset_nodes": "external_asset_graph_data",
        "resources": "external_resource_data",
        "asset_check_nodes": "external_asset_checks",
        "job_datas": "external_pipeline_datas",
        "job_refs": "external_job_refs",
    },
    skip_when_empty_fields={"pools"},
)
@record_custom
class RepositorySnap(IHaveNew):
    name: str
    schedules: Sequence["ScheduleSnap"]
    partition_sets: Sequence["PartitionSetSnap"]
    sensors: Sequence["SensorSnap"]
    asset_nodes: Sequence["AssetNodeSnap"]
    job_datas: Optional[Sequence["JobDataSnap"]]
    job_refs: Optional[Sequence["JobRefSnap"]]
    resources: Optional[Sequence["ResourceSnap"]]
    asset_check_nodes: Optional[Sequence["AssetCheckNodeSnap"]]
    metadata: Optional[MetadataMapping]
    utilized_env_vars: Optional[Mapping[str, Sequence["EnvVarConsumer"]]]

    def __new__(
        cls,
        name: str,
        schedules: Sequence["ScheduleSnap"],
        partition_sets: Sequence["PartitionSetSnap"],
        sensors: Optional[Sequence["SensorSnap"]] = None,
        asset_nodes: Optional[Sequence["AssetNodeSnap"]] = None,
        job_datas: Optional[Sequence["JobDataSnap"]] = None,
        job_refs: Optional[Sequence["JobRefSnap"]] = None,
        resources: Optional[Sequence["ResourceSnap"]] = None,
        asset_check_nodes: Optional[Sequence["AssetCheckNodeSnap"]] = None,
        metadata: Optional[MetadataMapping] = None,
        utilized_env_vars: Optional[Mapping[str, Sequence["EnvVarConsumer"]]] = None,
    ):
        return super().__new__(
            cls,
            name=name,
            schedules=schedules,
            partition_sets=partition_sets,
            sensors=sensors or [],
            asset_nodes=asset_nodes or [],
            job_datas=job_datas,
            job_refs=job_refs,
            resources=resources,
            asset_check_nodes=asset_check_nodes,
            metadata=metadata or {},
            utilized_env_vars=utilized_env_vars,
        )

    @classmethod
    def from_def(
        cls,
        repository_def: RepositoryDefinition,
        defer_snapshots: bool = False,
    ) -> Self:
        check.inst_param(repository_def, "repository_def", RepositoryDefinition)

        jobs = repository_def.get_all_jobs()
        if defer_snapshots:
            job_datas = None
            job_refs = sorted(
                [JobRefSnap.from_job_def(job) for job in jobs],
                key=lambda pd: pd.name,
            )
        else:
            job_datas = sorted(
                list(
                    map(
                        lambda job: JobDataSnap.from_job_def(job, include_parent_snapshot=True),
                        jobs,
                    )
                ),
                key=lambda pd: pd.name,
            )
            job_refs = None

        resource_datas = repository_def.get_top_level_resources()
        asset_node_snaps = asset_node_snaps_from_repo(repository_def)

        nested_resource_map = _get_nested_resources_map(
            resource_datas, repository_def.get_top_level_resources()
        )
        inverted_nested_resources_map: dict[str, dict[str, str]] = defaultdict(dict)
        for resource_key, nested_resources in nested_resource_map.items():
            for attribute, nested_resource in nested_resources.items():
                if nested_resource.type == NestedResourceType.TOP_LEVEL:
                    inverted_nested_resources_map[nested_resource.name][resource_key] = attribute

        resource_asset_usage_map: dict[str, list[AssetKey]] = defaultdict(list)
        # collect resource usage from normal non-source assets
        for asset in asset_node_snaps:
            if asset.required_top_level_resources:
                for resource_key in asset.required_top_level_resources:
                    resource_asset_usage_map[resource_key].append(asset.asset_key)

        resource_schedule_usage_map: dict[str, list[str]] = defaultdict(list)
        for schedule in repository_def.schedule_defs:
            if schedule.required_resource_keys:
                for resource_key in schedule.required_resource_keys:
                    resource_schedule_usage_map[resource_key].append(schedule.name)

        resource_sensor_usage_map: dict[str, list[str]] = defaultdict(list)
        for sensor in repository_def.sensor_defs:
            if sensor.required_resource_keys:
                for resource_key in sensor.required_resource_keys:
                    resource_sensor_usage_map[resource_key].append(sensor.name)

        resource_job_usage_map: ResourceJobUsageMap = _get_resource_job_usage(jobs)

        return cls(
            name=repository_def.name,
            schedules=sorted(
                [
                    ScheduleSnap.from_def(schedule_def, repository_def)
                    for schedule_def in repository_def.schedule_defs
                ],
                key=lambda sd: sd.name,
            ),
            # `PartitionSetDefinition` has been deleted, so we now construct `PartitionSetSnap`
            # from jobs instead of going through the intermediary `PartitionSetDefinition`. Eventually
            # we will remove `PartitionSetSnap` as well.
            partition_sets=sorted(
                [
                    PartitionSetSnap.from_job_def(job_def)
                    for job_def in repository_def.get_all_jobs()
                    if job_def.partitions_def is not None
                ],
                key=lambda pss: pss.name,
            ),
            sensors=sorted(
                [
                    SensorSnap.from_def(sensor_def, repository_def)
                    for sensor_def in repository_def.sensor_defs
                ],
                key=lambda sd: sd.name,
            ),
            asset_nodes=asset_node_snaps,
            job_datas=job_datas,
            job_refs=job_refs,
            resources=sorted(
                [
                    ResourceSnap.from_def(
                        res_data,
                        res_name,
                        nested_resource_map[res_name],
                        inverted_nested_resources_map[res_name],
                        resource_asset_usage_map,
                        resource_job_usage_map,
                        resource_schedule_usage_map,
                        resource_sensor_usage_map,
                    )
                    for res_name, res_data in resource_datas.items()
                ],
                key=lambda rd: rd.name,
            ),
            asset_check_nodes=asset_check_node_snaps_from_repo(repository_def),
            metadata=repository_def.metadata,
            utilized_env_vars={
                env_var: [
                    EnvVarConsumer(type=EnvVarConsumerType.RESOURCE, name=res_name)
                    for res_name in res_names
                ]
                for env_var, res_names in repository_def.get_env_vars_by_top_level_resource().items()
            },
        )

    def has_job_data(self):
        return self.job_datas is not None

    def get_job_datas(self) -> Sequence["JobDataSnap"]:
        if self.job_datas is None:
            check.failed("Snapshots were deferred, external_pipeline_data not loaded")
        return self.job_datas

    def get_job_refs(self) -> Sequence["JobRefSnap"]:
        if self.job_refs is None:
            check.failed("Snapshots were not deferred, job_refs not loaded")
        return self.job_refs

    def get_job_snap(self, name):
        check.str_param(name, "name")
        if self.job_datas is None:
            check.failed("Snapshots were deferred, external_pipeline_data not loaded")

        for job_data in self.job_datas:
            if job_data.name == name:
                return job_data.job

        check.failed("Could not find pipeline snapshot named " + name)

    def get_job_data(self, name):
        check.str_param(name, "name")
        if self.job_datas is None:
            check.failed("Snapshots were deferred, external_pipeline_data not loaded")

        for job_data in self.job_datas:
            if job_data.name == name:
                return job_data

        check.failed("Could not find external pipeline data named " + name)

    def get_schedule(self, name):
        check.str_param(name, "name")

        for schedule in self.schedules:
            if schedule.name == name:
                return schedule

        check.failed("Could not find external schedule data named " + name)

    def has_partition_set(self, name) -> bool:
        check.str_param(name, "name")
        for partition_set in self.partition_sets:
            if partition_set.name == name:
                return True

        return False

    def get_partition_set(self, name) -> "PartitionSetSnap":
        check.str_param(name, "name")

        for partition_set in self.partition_sets:
            if partition_set.name == name:
                return partition_set

        check.failed("Could not find external partition set data named " + name)

    def get_sensor(self, name):
        check.str_param(name, "name")

        for sensor in self.sensors:
            if sensor.name == name:
                return sensor

        check.failed("Could not find sensor data named " + name)


@whitelist_for_serdes(
    storage_name="ExternalPresetData", storage_field_names={"op_selection": "solid_selection"}
)
@record_custom
class PresetSnap(IHaveNew):
    name: str
    run_config: Mapping[str, object]
    op_selection: Optional[Sequence[str]]
    mode: str
    tags: Mapping[str, str]

    def __new__(
        cls,
        name: str,
        run_config: Optional[Mapping[str, object]],
        op_selection: Optional[Sequence[str]],
        mode: str,
        tags: Optional[Mapping[str, str]],
    ):
        return super().__new__(
            cls,
            name=name,
            run_config=run_config or {},
            op_selection=op_selection,
            mode=mode,
            tags=tags or {},
        )


_JOB_SNAP_STORAGE_FIELD = "pipeline_snapshot"


@whitelist_for_serdes(
    storage_name="ExternalPipelineData",
    storage_field_names={
        "job": _JOB_SNAP_STORAGE_FIELD,
        "parent_job": "parent_pipeline_snapshot",
    },
    # There was a period during which `JobDefinition` was a newer subclass of the legacy
    # `PipelineDefinition`, and `is_job` was a boolean field used to distinguish between the two
    # cases on this class.
    old_fields={"is_job": True},
)
@record
class JobDataSnap:
    name: str
    job: JobSnap
    active_presets: Sequence["PresetSnap"]
    parent_job: Optional[JobSnap]

    @classmethod
    def from_job_def(cls, job_def: JobDefinition, include_parent_snapshot: bool) -> Self:
        return cls(
            name=job_def.name,
            job=job_def.get_job_snapshot(),
            parent_job=job_def.get_parent_job_snapshot() if include_parent_snapshot else None,
            active_presets=active_presets_from_job_def(job_def),
        )


@whitelist_for_serdes(
    storage_name="ExternalPipelineSubsetResult",
    storage_field_names={"job_data_snap": "external_pipeline_data"},
)
@record
class RemoteJobSubsetResult:
    success: bool
    error: Optional[SerializableErrorInfo] = None
    job_data_snap: Optional[JobDataSnap] = None
    repository_python_origin: Optional[RepositoryPythonOrigin] = None


@whitelist_for_serdes
class EnvVarConsumerType(Enum):
    RESOURCE = "RESOURCE"


@whitelist_for_serdes
class EnvVarConsumer(NamedTuple):
    type: EnvVarConsumerType
    name: str


@whitelist_for_serdes
class NestedResourceType(Enum):
    ANONYMOUS = "ANONYMOUS"
    TOP_LEVEL = "TOP_LEVEL"


@whitelist_for_serdes
class NestedResource(NamedTuple):
    type: NestedResourceType
    name: str


@whitelist_for_serdes(storage_name="ExternalJobRef", old_fields={"is_legacy_pipeline": False})
@record
class JobRefSnap:
    name: str
    snapshot_id: str
    active_presets: Sequence["PresetSnap"]
    parent_snapshot_id: Optional[str]

    @classmethod
    def from_job_def(cls, job_def: JobDefinition) -> Self:
        check.inst_param(job_def, "job_def", JobDefinition)

        return cls(
            name=job_def.name,
            snapshot_id=job_def.get_job_snapshot_id(),
            parent_snapshot_id=None,
            active_presets=active_presets_from_job_def(job_def),
        )


@whitelist_for_serdes(
    storage_name="ExternalScheduleData",
    storage_field_names={"job_name": "pipeline_name", "op_selection": "solid_selection"},
    skip_when_empty_fields={"default_status"},
)
@record_custom
class ScheduleSnap(IHaveNew):
    name: str
    cron_schedule: Union[str, Sequence[str]]
    job_name: str
    op_selection: Optional[Sequence[str]]
    mode: Optional[str]
    environment_vars: Mapping[str, str]
    partition_set_name: Optional[str]
    execution_timezone: Optional[str]
    description: Optional[str]
    default_status: Optional[DefaultScheduleStatus]
    asset_selection: Optional[AssetSelection]
    tags: Mapping[str, str]
    metadata: Mapping[str, MetadataValue]

    def __new__(
        cls,
        name: str,
        cron_schedule: Union[str, Sequence[str]],
        job_name: str,
        op_selection: Optional[Sequence[str]],
        mode: Optional[str],
        environment_vars: Optional[Mapping[str, str]],
        partition_set_name: Optional[str],
        execution_timezone: Optional[str],
        description: Optional[str] = None,
        default_status: Optional[DefaultScheduleStatus] = None,
        asset_selection: Optional[AssetSelection] = None,
        tags: Optional[Mapping[str, str]] = None,
        metadata: Optional[Mapping[str, MetadataValue]] = None,
    ):
        if asset_selection is not None:
            check.invariant(
                is_whitelisted_for_serdes_object(asset_selection),
                "asset_selection must be serializable",
            )

        return super().__new__(
            cls,
            name=name,
            cron_schedule=cron_schedule,
            job_name=job_name,
            op_selection=op_selection,
            mode=mode,
            environment_vars=environment_vars or {},
            partition_set_name=partition_set_name,
            execution_timezone=execution_timezone,
            description=description,
            # Leave default_status as None if it's STOPPED to maintain stable back-compat IDs
            default_status=(
                DefaultScheduleStatus.RUNNING
                if default_status == DefaultScheduleStatus.RUNNING
                else None
            ),
            asset_selection=asset_selection,
            tags=tags or {},
            metadata=metadata or {},
        )

    @classmethod
    def from_def(
        cls, schedule_def: ScheduleDefinition, repository_def: RepositoryDefinition
    ) -> Self:
        if schedule_def.has_anonymous_job:
            job_def = check.inst(
                schedule_def.job,
                UnresolvedAssetJobDefinition,
                "Anonymous job should be UnresolvedAssetJobDefinition",
            )
            serializable_asset_selection = job_def.selection.to_serializable_asset_selection(
                repository_def.asset_graph
            )
        else:
            serializable_asset_selection = None

        return cls(
            name=schedule_def.name,
            cron_schedule=schedule_def.cron_schedule,
            job_name=schedule_def.job_name,
            op_selection=schedule_def.target.op_selection,
            mode=DEFAULT_MODE_NAME,
            environment_vars=schedule_def.environment_vars,
            partition_set_name=None,
            execution_timezone=schedule_def.execution_timezone,
            description=schedule_def.description,
            default_status=schedule_def.default_status,
            asset_selection=serializable_asset_selection,
            tags=schedule_def.tags,
            metadata=schedule_def.metadata,
        )


@whitelist_for_serdes(storage_name="ExternalScheduleExecutionErrorData")
@record
class ScheduleExecutionErrorSnap:
    error: Optional[SerializableErrorInfo]


@whitelist_for_serdes(
    storage_name="ExternalTargetData",
    storage_field_names={"job_name": "pipeline_name", "op_selection": "solid_selection"},
)
@record
class TargetSnap:
    job_name: str
    mode: str
    op_selection: Optional[Sequence[str]]


@whitelist_for_serdes(storage_name="ExternalSensorMetadata")
@record
class SensorMetadataSnap:
    """Stores sensor metadata which is available in the Dagster UI.

    This is an unfortunate legacy class that is out of line with our preferred pattern of storing
    standard `Mapping[str, MetadataValue]` under the metadata field. Because this class already
    existed when adding this standard metadata to sensors, we stash it on here as a field under
    `standard_metadata`.
    """

    asset_keys: Optional[Sequence[AssetKey]]
    standard_metadata: Optional[Mapping[str, MetadataValue]] = None


@whitelist_for_serdes(
    storage_name="ExternalSensorData",
    storage_field_names={"job_name": "pipeline_name", "op_selection": "solid_selection"},
    skip_when_empty_fields={"default_status", "sensor_type"},
)
@record_custom
class SensorSnap(IHaveNew):
    name: str
    job_name: Optional[str]
    op_selection: Optional[Sequence[str]]
    mode: Optional[str]
    min_interval: Optional[int]
    description: Optional[str]
    target_dict: Mapping[str, TargetSnap]
    metadata: Optional[SensorMetadataSnap]
    default_status: Optional[DefaultSensorStatus]
    sensor_type: Optional[SensorType]
    asset_selection: Optional[AssetSelection]
    tags: Mapping[str, str]
    run_tags: Mapping[str, str]

    def __new__(
        cls,
        name: str,
        job_name: Optional[str] = None,
        op_selection: Optional[Sequence[str]] = None,
        mode: Optional[str] = None,
        min_interval: Optional[int] = None,
        description: Optional[str] = None,
        target_dict: Optional[Mapping[str, TargetSnap]] = None,
        metadata: Optional[SensorMetadataSnap] = None,
        default_status: Optional[DefaultSensorStatus] = None,
        sensor_type: Optional[SensorType] = None,
        asset_selection: Optional[AssetSelection] = None,
        tags: Optional[Mapping[str, str]] = None,
        run_tags: Optional[Mapping[str, str]] = None,
    ):
        if job_name and not target_dict:
            # handle the legacy case where the ExternalSensorData was constructed from an earlier
            # version of dagster
            target_dict = {
                job_name: TargetSnap(
                    job_name=check.str_param(job_name, "job_name"),
                    mode=check.opt_str_param(mode, "mode", DEFAULT_MODE_NAME),
                    op_selection=check.opt_nullable_sequence_param(
                        op_selection, "op_selection", str
                    ),
                )
            }

        if asset_selection is not None:
            check.invariant(
                is_whitelisted_for_serdes_object(asset_selection),
                "asset_selection must be serializable",
            )

        return super().__new__(
            cls,
            name=name,
            job_name=job_name,  # keep legacy field populated
            op_selection=op_selection,  # keep legacy field populated
            mode=mode,  # keep legacy field populated
            min_interval=min_interval,
            description=description,
            target_dict=target_dict,
            metadata=metadata,
            # Leave default_status as None if it's STOPPED to maintain stable back-compat IDs
            default_status=(
                DefaultSensorStatus.RUNNING
                if default_status == DefaultSensorStatus.RUNNING
                else None
            ),
            sensor_type=sensor_type,
            asset_selection=asset_selection,
            tags=tags or {},
            run_tags=run_tags or {},
        )

    @classmethod
    def from_def(cls, sensor_def: SensorDefinition, repository_def: RepositoryDefinition) -> Self:
        first_target = sensor_def.targets[0] if sensor_def.targets else None

        asset_keys = None
        if isinstance(sensor_def, AssetSensorDefinition):
            asset_keys = [sensor_def.asset_key]

        if sensor_def.asset_selection is not None:
            target_dict = {
                base_asset_job_name: TargetSnap(
                    job_name=base_asset_job_name, mode=DEFAULT_MODE_NAME, op_selection=None
                )
                for base_asset_job_name in repository_def.get_implicit_asset_job_names()
            }

            serializable_asset_selection = (
                sensor_def.asset_selection.to_serializable_asset_selection(
                    repository_def.asset_graph
                )
            )
        else:
            target_dict = {
                target.job_name: TargetSnap(
                    job_name=target.job_name,
                    mode=DEFAULT_MODE_NAME,
                    op_selection=target.op_selection,
                )
                for target in sensor_def.targets
            }

            if sensor_def.has_anonymous_job:
                job_def = check.inst(
                    sensor_def.job,
                    UnresolvedAssetJobDefinition,
                    "Anonymous job should be UnresolvedAssetJobDefinition",
                )
                serializable_asset_selection = job_def.selection.to_serializable_asset_selection(
                    repository_def.asset_graph
                )
            else:
                serializable_asset_selection = None

        return cls(
            name=sensor_def.name,
            job_name=first_target.job_name if first_target else None,
            mode=None,
            op_selection=first_target.op_selection if first_target else None,
            target_dict=target_dict,
            min_interval=sensor_def.minimum_interval_seconds,
            description=sensor_def.description,
            metadata=SensorMetadataSnap(
                asset_keys=asset_keys, standard_metadata=sensor_def.metadata
            ),
            default_status=sensor_def.default_status,
            sensor_type=sensor_def.sensor_type,
            asset_selection=serializable_asset_selection,
            tags=sensor_def.tags,
            run_tags=(
                sensor_def.run_tags
                if isinstance(sensor_def, AutomationConditionSensorDefinition)
                else None
            ),
        )


@whitelist_for_serdes(storage_name="ExternalRepositoryErrorData")
@record
class RepositoryErrorSnap:
    error: Optional[SerializableErrorInfo]


@whitelist_for_serdes(storage_name="ExternalSensorExecutionErrorData")
@record
class SensorExecutionErrorSnap:
    error: Optional[SerializableErrorInfo]


@whitelist_for_serdes(storage_name="ExternalExecutionParamsData")
@record_custom
class ExecutionParamsSnap(IHaveNew):
    run_config: Mapping[str, object]
    tags: Mapping[str, str]

    def __new__(
        cls,
        run_config: Optional[Mapping[str, object]] = None,
        tags: Optional[Mapping[str, str]] = None,
    ):
        return super().__new__(
            cls,
            run_config=run_config or {},
            tags=tags or {},
        )


@whitelist_for_serdes(storage_name="ExternalExecutionParamsErrorData")
@record
class ExecutionParamsErrorSnap:
    error: Optional[SerializableErrorInfo]


class PartitionsSnap(ABC):
    @classmethod
    def from_def(cls, partitions_def: PartitionsDefinition) -> "PartitionsSnap":
        if isinstance(partitions_def, TimeWindowPartitionsDefinition):
            return TimeWindowPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, StaticPartitionsDefinition):
            return StaticPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, MultiPartitionsDefinition):
            return MultiPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, DynamicPartitionsDefinition):
            return DynamicPartitionsSnap.from_def(partitions_def)
        else:
            raise DagsterInvalidDefinitionError(
                "Only static, time window, multi-dimensional partitions, and dynamic partitions"
                " definitions with a name parameter are currently supported."
            )

    @abstractmethod
    def get_partitions_definition(self) -> PartitionsDefinition: ...


@whitelist_for_serdes(storage_name="ExternalTimeWindowPartitionsDefinitionData")
@record
class TimeWindowPartitionsSnap(PartitionsSnap):
    start: float
    timezone: Optional[str]
    fmt: str
    end_offset: int
    end: Optional[float] = None
    cron_schedule: Optional[str] = None
    # superseded by cron_schedule, but kept around for backcompat
    schedule_type: Optional[ScheduleType] = None
    # superseded by cron_schedule, but kept around for backcompat
    minute_offset: Optional[int] = None
    # superseded by cron_schedule, but kept around for backcompat
    hour_offset: Optional[int] = None
    # superseded by cron_schedule, but kept around for backcompat
    day_offset: Optional[int] = None

    @classmethod
    def from_def(cls, partitions_def: TimeWindowPartitionsDefinition) -> Self:
        check.inst_param(partitions_def, "partitions_def", TimeWindowPartitionsDefinition)
        return cls(
            cron_schedule=partitions_def.cron_schedule,
            start=partitions_def.start.timestamp(),
            end=partitions_def.end.timestamp() if partitions_def.end else None,
            timezone=partitions_def.timezone,
            fmt=partitions_def.fmt,
            end_offset=partitions_def.end_offset,
        )

    def get_partitions_definition(self):
        if self.cron_schedule is not None:
            return TimeWindowPartitionsDefinition(
                cron_schedule=self.cron_schedule,
                start=datetime_from_timestamp(self.start, tz=self.timezone),  # pyright: ignore[reportArgumentType]
                timezone=self.timezone,
                fmt=self.fmt,
                end_offset=self.end_offset,
                end=(datetime_from_timestamp(self.end, tz=self.timezone) if self.end else None),  # pyright: ignore[reportArgumentType]
            )
        else:
            # backcompat case
            return TimeWindowPartitionsDefinition(
                schedule_type=self.schedule_type,
                start=datetime_from_timestamp(self.start, tz=self.timezone),  # pyright: ignore[reportArgumentType]
                timezone=self.timezone,
                fmt=self.fmt,
                end_offset=self.end_offset,
                end=(datetime_from_timestamp(self.end, tz=self.timezone) if self.end else None),  # pyright: ignore[reportArgumentType]
                minute_offset=self.minute_offset,
                hour_offset=self.hour_offset,
                day_offset=self.day_offset,
            )


def _dedup_partition_keys(keys: Sequence[str]) -> Sequence[str]:
    # Use both a set and a list here to preserve lookup performance in case of large inputs. (We
    # can't just use a set because we need to preserve ordering.)
    seen_keys: set[str] = set()
    new_keys: list[str] = []
    for key in keys:
        if key not in seen_keys:
            new_keys.append(key)
            seen_keys.add(key)
    return new_keys


@whitelist_for_serdes(storage_name="ExternalStaticPartitionsDefinitionData")
@record_custom(checked=False)
class StaticPartitionsSnap(PartitionsSnap, IHaveNew):
    partition_keys: Sequence[str]

    def __new__(cls, partition_keys: Sequence[str]):
        # for back compat reasons we allow str as a Sequence[str] here
        if not isinstance(partition_keys, str):
            check.sequence_param(
                partition_keys,
                "partition_keys",
                of_type=str,
            )

        return super().__new__(
            cls,
            partition_keys=partition_keys,
        )

    @classmethod
    def from_def(cls, partitions_def: StaticPartitionsDefinition) -> Self:
        check.inst_param(partitions_def, "partitions_def", StaticPartitionsDefinition)
        return cls(partition_keys=partitions_def.get_partition_keys())

    def get_partitions_definition(self):
        # v1.4 made `StaticPartitionsDefinition` error if given duplicate keys. This caused
        # host process errors for users who had not upgraded their user code to 1.4 and had dup
        # keys, since the host process `StaticPartitionsDefinition` would throw an error.
        keys = _dedup_partition_keys(self.partition_keys)
        return StaticPartitionsDefinition(keys)


@whitelist_for_serdes(
    storage_name="ExternalPartitionDimensionDefinition",
    storage_field_names={"partitions": "external_partitions_def_data"},
)
@record
class PartitionDimensionSnap:
    name: str
    partitions: PartitionsSnap


@whitelist_for_serdes(
    storage_name="ExternalMultiPartitionsDefinitionData",
    storage_field_names={"partition_dimensions": "external_partition_dimension_definitions"},
)
@record
class MultiPartitionsSnap(PartitionsSnap):
    partition_dimensions: Sequence[PartitionDimensionSnap]

    @classmethod
    def from_def(cls, partitions_def: MultiPartitionsDefinition) -> Self:
        check.inst_param(partitions_def, "partitions_def", MultiPartitionsDefinition)

        return cls(
            partition_dimensions=[
                PartitionDimensionSnap(
                    name=dimension.name,
                    partitions=PartitionsSnap.from_def(dimension.partitions_def),
                )
                for dimension in partitions_def.partitions_defs
            ]
        )

    def get_partitions_definition(self):
        return MultiPartitionsDefinition(
            {
                partition_dimension.name: (
                    partition_dimension.partitions.get_partitions_definition()
                )
                for partition_dimension in self.partition_dimensions
            }
        )


@whitelist_for_serdes(storage_name="ExternalDynamicPartitionsDefinitionData")
@record
class DynamicPartitionsSnap(PartitionsSnap):
    name: str

    @classmethod
    def from_def(cls, partitions_def: DynamicPartitionsDefinition) -> Self:
        check.inst_param(partitions_def, "partitions_def", DynamicPartitionsDefinition)
        if partitions_def.name is None:
            raise DagsterInvalidDefinitionError(
                "Dagster does not support dynamic partitions definitions without a name parameter."
            )
        return cls(name=partitions_def.name)

    def get_partitions_definition(self):
        return DynamicPartitionsDefinition(name=self.name)


@whitelist_for_serdes(
    storage_name="ExternalPartitionSetData",
    storage_field_names={
        "job_name": "pipeline_name",
        "op_selection": "solid_selection",
        "partitions": "external_partitions_data",
    },
)
@record
class PartitionSetSnap:
    name: str
    job_name: str
    op_selection: Optional[Sequence[str]]
    mode: Optional[str]
    partitions: Optional[PartitionsSnap] = None
    backfill_policy: Optional[BackfillPolicy] = None

    @classmethod
    def from_job_def(cls, job_def: JobDefinition) -> Self:
        check.inst_param(job_def, "job_def", JobDefinition)
        partitions_def = check.not_none(job_def.partitions_def)

        partitions_snap: Optional[PartitionsSnap] = None
        if isinstance(partitions_def, TimeWindowPartitionsDefinition):
            partitions_snap = TimeWindowPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, StaticPartitionsDefinition):
            partitions_snap = StaticPartitionsSnap.from_def(partitions_def)
        elif (
            isinstance(partitions_def, DynamicPartitionsDefinition)
            and partitions_def.name is not None
        ):
            partitions_snap = DynamicPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, MultiPartitionsDefinition):
            partitions_snap = MultiPartitionsSnap.from_def(partitions_def)
        else:
            partitions_snap = None

        return cls(
            name=partition_set_snap_name_for_job_name(job_def.name),
            job_name=job_def.name,
            op_selection=None,
            mode=DEFAULT_MODE_NAME,
            partitions=partitions_snap,
            backfill_policy=job_def.backfill_policy,
        )


@whitelist_for_serdes(storage_name="ExternalPartitionNamesData")
@record_custom
class PartitionNamesSnap(IHaveNew):
    partition_names: Sequence[str]

    def __new__(cls, partition_names: Optional[Sequence[str]] = None):
        return super().__new__(
            cls,
            partition_names=partition_names or [],
        )


@whitelist_for_serdes(storage_name="ExternalPartitionConfigData")
@record_custom
class PartitionConfigSnap(IHaveNew):
    name: str
    run_config: Mapping[str, object]

    def __new__(cls, name: str, run_config: Optional[Mapping[str, object]] = None):
        return super().__new__(
            cls,
            name=name,
            run_config=run_config or {},
        )


@whitelist_for_serdes(storage_name="ExternalPartitionTagsData")
@record_custom
class PartitionTagsSnap(IHaveNew):
    name: str
    tags: Mapping[str, object]

    def __new__(cls, name: str, tags: Optional[Mapping[str, str]] = None):
        return super().__new__(
            cls,
            name=name,
            tags=tags or {},
        )


@whitelist_for_serdes(storage_name="ExternalPartitionExecutionParamData")
@record
class PartitionExecutionParamSnap:
    name: str
    tags: Mapping[str, str]
    run_config: Mapping[str, object]


@whitelist_for_serdes(storage_name="ExternalPartitionSetExecutionParamData")
@record
class PartitionSetExecutionParamSnap:
    partition_data: Sequence[PartitionExecutionParamSnap]


@whitelist_for_serdes(storage_name="ExternalPartitionExecutionErrorData")
@record
class PartitionExecutionErrorSnap:
    error: Optional[SerializableErrorInfo]


@whitelist_for_serdes(
    storage_name="ExternalAssetDependency",
    storage_field_names={"parent_asset_key": "upstream_asset_key"},
)
@record
class AssetParentEdgeSnap:
    """A definition of a directed edge in the logical asset graph.

    An upstream asset that's depended on, and the corresponding input name in the downstream asset
    that depends on it.
    """

    parent_asset_key: AssetKey
    input_name: Optional[str] = None
    output_name: Optional[str] = None
    partition_mapping: Optional[PartitionMapping] = None


@whitelist_for_serdes(
    storage_name="ExternalAssetDependedBy",
    storage_field_names={"child_asset_key": "downstream_asset_key"},
)
@record
class AssetChildEdgeSnap:
    """A definition of a directed edge in the logical asset graph.

    An downstream asset that's depended by, and the corresponding input name in the upstream
    asset that it depends on.
    """

    child_asset_key: AssetKey
    input_name: Optional[str] = None
    output_name: Optional[str] = None


@whitelist_for_serdes(storage_name="ExternalResourceConfigEnvVar")
@record
class ResourceConfigEnvVarSnap:
    name: str


ResourceValueSnap: TypeAlias = Union[str, ResourceConfigEnvVarSnap]


UNKNOWN_RESOURCE_TYPE = "Unknown"


@whitelist_for_serdes
@record
class ResourceJobUsageEntry:
    """Stores information about where a resource is used in a job."""

    job_name: str
    node_handles: list[NodeHandle]


@whitelist_for_serdes(storage_name="ExternalResourceData")
@record_custom
class ResourceSnap(IHaveNew):
    """Serializable data associated with a top-level resource in a Repository, e.g. one bound using the Definitions API.

    Includes information about the resource definition and config schema, user-passed values, etc.
    """

    name: str
    resource_snapshot: ResourceDefSnap
    configured_values: dict[str, ResourceValueSnap]
    config_field_snaps: list[ConfigFieldSnap]
    config_schema_snap: ConfigSchemaSnapshot
    nested_resources: dict[str, NestedResource]
    parent_resources: dict[str, str]
    resource_type: str
    is_top_level: bool
    asset_keys_using: list[AssetKey]
    job_ops_using: list[ResourceJobUsageEntry]
    dagster_maintained: bool
    schedules_using: list[str]
    sensors_using: list[str]

    def __new__(
        cls,
        name: str,
        resource_snapshot: ResourceDefSnap,
        configured_values: Mapping[str, ResourceValueSnap],
        config_field_snaps: Sequence[ConfigFieldSnap],
        config_schema_snap: ConfigSchemaSnapshot,
        nested_resources: Optional[Mapping[str, NestedResource]] = None,
        parent_resources: Optional[Mapping[str, str]] = None,
        resource_type: str = UNKNOWN_RESOURCE_TYPE,
        is_top_level: bool = True,
        asset_keys_using: Optional[Sequence[AssetKey]] = None,
        job_ops_using: Optional[Sequence[ResourceJobUsageEntry]] = None,
        dagster_maintained: bool = False,
        schedules_using: Optional[Sequence[str]] = None,
        sensors_using: Optional[Sequence[str]] = None,
    ):
        return super().__new__(
            cls,
            name=name,
            resource_snapshot=resource_snapshot,
            configured_values=dict(
                check.mapping_param(
                    configured_values,
                    "configured_values",
                    key_type=str,
                    value_type=(str, ResourceConfigEnvVarSnap),
                )
            ),
            config_field_snaps=config_field_snaps,
            config_schema_snap=config_schema_snap,
            nested_resources=dict(
                check.opt_mapping_param(
                    nested_resources, "nested_resources", key_type=str, value_type=NestedResource
                )
            ),
            parent_resources=dict(
                check.opt_mapping_param(
                    parent_resources, "parent_resources", key_type=str, value_type=str
                )
            ),
            is_top_level=is_top_level,
            resource_type=resource_type,
            asset_keys_using=list(
                check.opt_sequence_param(asset_keys_using, "asset_keys_using", of_type=AssetKey)
            ),
            job_ops_using=list(
                check.opt_sequence_param(
                    job_ops_using, "job_ops_using", of_type=ResourceJobUsageEntry
                )
            ),
            dagster_maintained=dagster_maintained,
            schedules_using=list(
                check.opt_sequence_param(schedules_using, "schedules_using", of_type=str)
            ),
            sensors_using=list(
                check.opt_sequence_param(sensors_using, "sensors_using", of_type=str)
            ),
        )

    @classmethod
    def from_def(
        cls,
        resource_def: ResourceDefinition,
        name: str,
        nested_resources: Mapping[str, NestedResource],
        parent_resources: Mapping[str, str],
        resource_asset_usage_map: Mapping[str, list[AssetKey]],
        resource_job_usage_map: "ResourceJobUsageMap",
        resource_schedule_usage_map: Mapping[str, list[str]],
        resource_sensor_usage_map: Mapping[str, list[str]],
    ) -> Self:
        check.inst_param(resource_def, "resource_def", ResourceDefinition)

        # Once values on a resource object are bound, the config schema for those fields is no
        # longer visible. We walk up the list of parent schemas to find the base, unconfigured
        # schema so we can display all fields in the UI.
        unconfigured_config_schema = resource_def.config_schema
        while (
            isinstance(unconfigured_config_schema, ConfiguredDefinitionConfigSchema)
            and unconfigured_config_schema.parent_def.config_schema
        ):
            unconfigured_config_schema = unconfigured_config_schema.parent_def.config_schema

        config_type = check.not_none(unconfigured_config_schema.config_type)
        unconfigured_config_type_snap = snap_from_config_type(config_type)

        config_schema_default = cast(
            Mapping[str, Any],
            (
                json.loads(resource_def.config_schema.default_value_as_json_str)
                if resource_def.config_schema.default_provided
                else {}
            ),
        )

        # Right now, .configured sets the default value of the top-level Field
        # we parse the JSON and break it out into defaults for each individual nested Field
        # for display in the UI
        configured_values = {
            k: resource_value_snap_from_raw(v) for k, v in config_schema_default.items()
        }

        resource_type_def = resource_def

        # use the resource function name as the resource type if it's a function resource
        # (ie direct instantiation of ResourceDefinition or IOManagerDefinition)
        if type(resource_type_def) in (ResourceDefinition, IOManagerDefinition):
            original_resource_fn = (
                resource_type_def._hardcoded_resource_type  # noqa: SLF001
                if resource_type_def._hardcoded_resource_type  # noqa: SLF001
                else resource_type_def.resource_fn
            )
            module_name = check.not_none(inspect.getmodule(original_resource_fn)).__name__
            resource_type = f"{module_name}.{original_resource_fn.__name__}"
        # if it's a Pythonic resource, get the underlying Pythonic class name
        elif isinstance(
            resource_type_def,
            (
                ConfigurableResourceFactoryResourceDefinition,
                ConfigurableIOManagerFactoryResourceDefinition,
            ),
        ):
            resource_type = _get_class_name(resource_type_def.configurable_resource_cls)
        else:
            resource_type = _get_class_name(type(resource_type_def))

        dagster_maintained = (
            resource_type_def._is_dagster_maintained()  # noqa: SLF001
            if type(resource_type_def)
            in (
                ResourceDefinition,
                IOManagerDefinition,
                ConfigurableResourceFactoryResourceDefinition,
                ConfigurableIOManagerFactoryResourceDefinition,
            )
            else False
        )

        return cls(
            name=name,
            resource_snapshot=build_resource_def_snap(name, resource_def),
            configured_values=configured_values,
            config_field_snaps=unconfigured_config_type_snap.fields or [],
            config_schema_snap=config_type.schema_snapshot,
            nested_resources=nested_resources,
            parent_resources=parent_resources,
            is_top_level=True,
            asset_keys_using=resource_asset_usage_map.get(name, []),
            job_ops_using=resource_job_usage_map.get(name, []),
            schedules_using=resource_schedule_usage_map.get(name, []),
            sensors_using=resource_sensor_usage_map.get(name, []),
            resource_type=resource_type,
            dagster_maintained=dagster_maintained,
        )


@whitelist_for_serdes(
    storage_name="ExternalAssetCheck",
    storage_field_names={"execution_set_identifier": "atomic_execution_unit_id"},
)
@record_custom
class AssetCheckNodeSnap(IHaveNew):
    """Serializable data associated with an asset check."""

    name: str
    asset_key: AssetKey
    description: Optional[str]
    execution_set_identifier: Optional[str]
    job_names: Sequence[str]
    blocking: bool
    additional_asset_keys: Sequence[AssetKey]
    automation_condition: Optional[AutomationCondition]
    automation_condition_snapshot: Optional[AutomationConditionSnapshot]

    def __new__(
        cls,
        name: str,
        asset_key: AssetKey,
        description: Optional[str],
        execution_set_identifier: Optional[str] = None,
        job_names: Optional[Sequence[str]] = None,
        blocking: bool = False,
        additional_asset_keys: Optional[Sequence[AssetKey]] = None,
        automation_condition: Optional[AutomationCondition] = None,
        automation_condition_snapshot: Optional[AutomationConditionSnapshot] = None,
    ):
        return super().__new__(
            cls,
            name=name,
            asset_key=asset_key,
            description=description,
            execution_set_identifier=execution_set_identifier,
            job_names=job_names or [],
            blocking=blocking,
            additional_asset_keys=additional_asset_keys or [],
            automation_condition=automation_condition,
            automation_condition_snapshot=automation_condition_snapshot,
        )

    @property
    def key(self) -> AssetCheckKey:
        return AssetCheckKey(asset_key=self.asset_key, name=self.name)


class BackcompatTeamOwnerFieldDeserializer(FieldSerializer):
    """Up through Dagster 1.7.7, asset owners provided as "team:foo" would be serialized as "foo"
    going forward, they're serialized as "team:foo".
    """

    def unpack(self, __unpacked_value, whitelist_map, context):
        return (
            [
                owner if (is_valid_email(owner) or owner.startswith("team:")) else f"team:{owner}"
                for owner in cast(Sequence[str], __unpacked_value)
            ]
            if __unpacked_value is not None
            else None
        )

    def pack(self, __unpacked_value, whitelist_map, descent_path):
        return __unpacked_value


@whitelist_for_serdes(
    storage_name="ExternalAssetNode",
    storage_field_names={
        "child_edges": "depended_by",
        "parent_edges": "dependencies",
        "metadata": "metadata_entries",
        "execution_set_identifier": "atomic_execution_unit_id",
        "description": "op_description",
        "partitions": "partitions_def_data",
    },
    field_serializers={
        "metadata": MetadataFieldSerializer,
        "owners": BackcompatTeamOwnerFieldDeserializer,
    },
)
@suppress_dagster_warnings
@record_custom
class AssetNodeSnap(IHaveNew):
    """A definition of a node in the logical asset graph.

    A function for computing the asset and an identifier for that asset.
    """

    asset_key: AssetKey
    parent_edges: Sequence[AssetParentEdgeSnap]
    child_edges: Sequence[AssetChildEdgeSnap]
    execution_type: AssetExecutionType
    pools: set[str]
    compute_kind: Optional[str]
    op_name: Optional[str]
    op_names: Sequence[str]
    code_version: Optional[str]
    node_definition_name: Optional[str]
    graph_name: Optional[str]
    description: Optional[str]
    job_names: Sequence[str]
    partitions: Optional[PartitionsSnap]
    output_name: Optional[str]
    metadata: Mapping[str, MetadataValue]
    tags: Optional[Mapping[str, str]]
    group_name: str
    freshness_policy: Optional[FreshnessPolicy]
    is_source: bool
    is_observable: bool
    # If a set of assets can't be materialized independently from each other, they will all
    # have the same execution_set_identifier. This ID should be stable across reloads and
    # unique deployment-wide.
    execution_set_identifier: Optional[str]
    required_top_level_resources: Optional[Sequence[str]]
    auto_materialize_policy: Optional[AutoMaterializePolicy]
    automation_condition_snapshot: Optional[AutomationConditionSnapshot]
    backfill_policy: Optional[BackfillPolicy]
    auto_observe_interval_minutes: Optional[Union[float, int]]
    owners: Optional[Sequence[str]]

    def __new__(
        cls,
        asset_key: AssetKey,
        parent_edges: Sequence[AssetParentEdgeSnap],
        child_edges: Sequence[AssetChildEdgeSnap],
        execution_type: Optional[AssetExecutionType] = None,
        pools: Optional[set[str]] = None,
        compute_kind: Optional[str] = None,
        op_name: Optional[str] = None,
        op_names: Optional[Sequence[str]] = None,
        code_version: Optional[str] = None,
        node_definition_name: Optional[str] = None,
        graph_name: Optional[str] = None,
        description: Optional[str] = None,
        job_names: Optional[Sequence[str]] = None,
        partitions: Optional[PartitionsSnap] = None,
        output_name: Optional[str] = None,
        metadata: Optional[Mapping[str, MetadataValue]] = None,
        tags: Optional[Mapping[str, str]] = None,
        group_name: Optional[str] = None,
        freshness_policy: Optional[FreshnessPolicy] = None,
        is_source: Optional[bool] = None,
        is_observable: bool = False,
        execution_set_identifier: Optional[str] = None,
        required_top_level_resources: Optional[Sequence[str]] = None,
        auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
        automation_condition_snapshot: Optional[AutomationConditionSnapshot] = None,
        backfill_policy: Optional[BackfillPolicy] = None,
        auto_observe_interval_minutes: Optional[Union[float, int]] = None,
        owners: Optional[Sequence[str]] = None,
    ):
        metadata = normalize_metadata(
            check.opt_mapping_param(metadata, "metadata", key_type=str), allow_invalid=True
        )

        # backcompat logic for execution type specified via metadata
        if SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE in metadata:
            val = metadata[SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE]
            if not isinstance(val, TextMetadataValue):
                check.failed(
                    f"Expected metadata value for key {SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE} to be a TextMetadataValue, got {val}"
                )
            metadata_execution_type = AssetExecutionType[check.not_none(val.value)]
            if execution_type is not None:
                check.invariant(
                    execution_type == metadata_execution_type,
                    f"Execution type {execution_type} in metadata does not match type inferred from metadata {metadata_execution_type}",
                )
            execution_type = metadata_execution_type
        else:
            if is_source and is_observable:
                default_execution_type = AssetExecutionType.OBSERVATION
            elif is_source:
                default_execution_type = AssetExecutionType.UNEXECUTABLE
            else:
                default_execution_type = AssetExecutionType.MATERIALIZATION

            execution_type = (
                check.opt_inst_param(
                    execution_type,
                    "execution_type",
                    AssetExecutionType,
                )
                or default_execution_type
            )

        # backcompat logic to handle AssetNodeSnaps serialized without op_names/graph_name
        if not op_names:
            op_names = list(filter(None, [op_name]))

        # backcompat logic to handle AssetNodeSnaps serialzied without is_source
        if is_source is None:
            # prior to this field being added, all non-source assets must be part of at least one
            # job, and no source assets could be part of any job
            is_source = len(job_names or []) == 0

        return super().__new__(
            cls,
            asset_key=asset_key,
            parent_edges=parent_edges or [],
            child_edges=child_edges or [],
            compute_kind=compute_kind,
            pools=pools or set(),
            op_name=op_name,
            op_names=op_names or [],
            code_version=code_version,
            node_definition_name=node_definition_name,
            graph_name=graph_name,
            description=description,
            job_names=job_names or [],
            partitions=partitions,
            output_name=output_name,
            metadata=metadata,
            tags=tags or {},
            # Newer code always passes a string group name when constructing these, but we assign
            # the default here for backcompat.
            group_name=group_name or DEFAULT_GROUP_NAME,
            freshness_policy=freshness_policy,
            is_source=is_source,
            is_observable=is_observable,
            execution_set_identifier=execution_set_identifier,
            required_top_level_resources=required_top_level_resources or [],
            auto_materialize_policy=auto_materialize_policy,
            automation_condition_snapshot=automation_condition_snapshot,
            backfill_policy=backfill_policy,
            auto_observe_interval_minutes=auto_observe_interval_minutes,
            owners=owners or [],
            execution_type=execution_type,
        )

    @property
    def is_materializable(self) -> bool:
        return self.execution_type == AssetExecutionType.MATERIALIZATION

    @property
    def is_external(self) -> bool:
        return self.execution_type != AssetExecutionType.MATERIALIZATION

    @property
    def is_executable(self) -> bool:
        return self.execution_type != AssetExecutionType.UNEXECUTABLE

    @property
    def automation_condition(self) -> Optional[AutomationCondition]:
        if self.auto_materialize_policy is not None:
            return self.auto_materialize_policy.to_automation_condition()
        else:
            return None


ResourceJobUsageMap: TypeAlias = dict[str, list[ResourceJobUsageEntry]]


class NodeHandleResourceUse(NamedTuple):
    resource_key: str
    node_handle: NodeHandle


def _get_resource_usage_from_node(
    pipeline: JobDefinition,
    node: Node,
    parent_handle: Optional[NodeHandle] = None,
) -> Iterable[NodeHandleResourceUse]:
    handle = NodeHandle(node.name, parent_handle)
    if isinstance(node, OpNode):
        for resource_req in node.get_resource_requirements(pipeline.graph):
            if isinstance(resource_req, ResourceKeyRequirement):
                yield NodeHandleResourceUse(resource_req.key, handle)
    elif isinstance(node, GraphNode):
        for nested_node in node.definition.nodes:
            yield from _get_resource_usage_from_node(pipeline, nested_node, handle)


def _get_resource_job_usage(job_defs: Sequence[JobDefinition]) -> ResourceJobUsageMap:
    resource_job_usage_map: dict[str, list[ResourceJobUsageEntry]] = defaultdict(list)

    for job_def in job_defs:
        job_name = job_def.name
        if is_reserved_asset_job_name(job_name):
            continue

        resource_usage: list[NodeHandleResourceUse] = []
        for solid in job_def.nodes_in_topological_order:
            resource_usage += [use for use in _get_resource_usage_from_node(job_def, solid)]
        node_use_by_key: dict[str, list[NodeHandle]] = defaultdict(list)
        for use in resource_usage:
            node_use_by_key[use.resource_key].append(use.node_handle)
        for resource_key in node_use_by_key:
            resource_job_usage_map[resource_key].append(
                ResourceJobUsageEntry(
                    job_name=job_def.name, node_handles=node_use_by_key[resource_key]
                )
            )

    return resource_job_usage_map


def asset_check_node_snaps_from_repo(repo: RepositoryDefinition) -> Sequence[AssetCheckNodeSnap]:
    job_names_by_check_key: dict[AssetCheckKey, list[str]] = defaultdict(list)

    for job_def in repo.get_all_jobs():
        asset_layer = job_def.asset_layer
        for check_key in asset_layer.asset_graph.asset_check_keys:
            job_names_by_check_key[check_key].append(job_def.name)

    asset_check_node_snaps: list[AssetCheckNodeSnap] = []
    for check_key, job_names in job_names_by_check_key.items():
        spec = repo.asset_graph.get_check_spec(check_key)
        automation_condition, automation_condition_snapshot = resolve_automation_condition_args(
            spec.automation_condition
        )
        asset_check_node_snaps.append(
            AssetCheckNodeSnap(
                name=check_key.name,
                asset_key=check_key.asset_key,
                description=spec.description,
                execution_set_identifier=repo.asset_graph.get_execution_set_identifier(check_key),
                job_names=job_names,
                blocking=spec.blocking,
                additional_asset_keys=[dep.asset_key for dep in spec.additional_deps],
                automation_condition=automation_condition,
                automation_condition_snapshot=automation_condition_snapshot,
            )
        )

    return sorted(asset_check_node_snaps, key=lambda check: (check.asset_key, check.name))


def asset_node_snaps_from_repo(repo: RepositoryDefinition) -> Sequence[AssetNodeSnap]:
    # First iterate over all job defs to identify a "primary node" for each materializable asset
    # key. This is the node that will be used to populate the AssetNodeSnap. We need to identify
    # a primary node because the same asset can be materialized as part of multiple jobs.
    primary_node_pairs_by_asset_key: dict[AssetKey, tuple[NodeOutputHandle, JobDefinition]] = {}
    job_defs_by_asset_key: dict[AssetKey, list[JobDefinition]] = {}
    for job_def in repo.get_all_jobs():
        asset_layer = job_def.asset_layer
        asset_keys_by_node_output = asset_layer.asset_keys_by_node_output_handle
        for node_output_handle, asset_key in asset_keys_by_node_output.items():
            if asset_key not in asset_layer.asset_keys_for_node(node_output_handle.node_handle):
                continue
            if asset_key not in primary_node_pairs_by_asset_key:
                primary_node_pairs_by_asset_key[asset_key] = (node_output_handle, job_def)
            job_defs_by_asset_key.setdefault(asset_key, []).append(job_def)

    asset_node_snaps: list[AssetNodeSnap] = []
    asset_graph = repo.asset_graph
    for key in sorted(asset_graph.get_all_asset_keys()):
        asset_node = asset_graph.get(key)

        # Materializable assets (which are always part of at least one job, due to asset base jobs)
        # have various fields related to their op/output/jobs etc defined. External assets have null
        # values for all these fields.
        if key in primary_node_pairs_by_asset_key:
            output_handle, job_def = primary_node_pairs_by_asset_key[key]

            root_node_handle = output_handle.node_handle.root
            node_def = job_def.graph.get_node(output_handle.node_handle).definition
            node_handles = job_def.asset_layer.upstream_dep_op_handles(key)

            # graph_name is only set for assets that are produced by nested ops.
            graph_name = (
                root_node_handle.name if root_node_handle != output_handle.node_handle else None
            )
            op_defs = [
                cast(OpDefinition, job_def.graph.get_node(node_handle).definition)
                for node_handle in node_handles
                if isinstance(job_def.graph.get_node(node_handle).definition, OpDefinition)
            ]
            pools = {op_def.pool for op_def in op_defs if op_def.pool}
            op_names = sorted([str(handle) for handle in node_handles])
            op_name = graph_name or next(iter(op_names), None) or node_def.name
            job_names = sorted([jd.name for jd in job_defs_by_asset_key[key]])
            compute_kind = node_def.tags.get(COMPUTE_KIND_TAG)
            node_definition_name = node_def.name

            # Confusingly, the `name` field sometimes mismatches the `name` field on the
            # OutputDefinition. We need to fix this.
            output_name = node_def.output_def_named(output_handle.output_name).name
            required_top_level_resources = (
                sorted(node_def.required_resource_keys)
                if isinstance(node_def, OpDefinition)
                else []
            )

        else:
            graph_name = None
            pools = set()
            op_names = []
            op_name = None
            job_names = []
            compute_kind = None
            node_definition_name = None
            output_name = None
            required_top_level_resources = []

        # Partition mappings are only exposed on the AssetNodeSnap if at least one asset is
        # partitioned and the partition mapping is one of the builtin types.
        partition_mappings: dict[AssetKey, Optional[PartitionMapping]] = {}
        builtin_partition_mapping_types = get_builtin_partition_mapping_types()
        for pk in asset_node.parent_keys:
            # directly access the partition mapping to avoid the inference step of
            # get_partition_mapping, as we want to defer the inference to the global RemoteAssetGraph
            partition_mapping = repo.asset_graph.get(key).partition_mappings.get(pk)
            if (
                asset_node.partitions_def or repo.asset_graph.get(pk).partitions_def
            ) and isinstance(partition_mapping, builtin_partition_mapping_types):
                partition_mappings[pk] = partition_mapping

        automation_condition, automation_condition_snapshot = resolve_automation_condition_args(
            asset_node.automation_condition
        )
        asset_node_snaps.append(
            AssetNodeSnap(
                asset_key=key,
                parent_edges=[
                    AssetParentEdgeSnap(
                        parent_asset_key=pk, partition_mapping=partition_mappings.get(pk)
                    )
                    for pk in sorted(asset_node.parent_keys)
                ],
                child_edges=[
                    AssetChildEdgeSnap(child_asset_key=k) for k in sorted(asset_node.child_keys)
                ],
                execution_type=asset_node.execution_type,
                compute_kind=compute_kind,
                pools=pools,
                op_name=op_name,
                op_names=op_names,
                code_version=asset_node.code_version,
                node_definition_name=node_definition_name,
                graph_name=graph_name,
                description=asset_node.description,
                job_names=job_names,
                partitions=(
                    PartitionsSnap.from_def(asset_node.partitions_def)
                    if asset_node.partitions_def
                    else None
                ),
                output_name=output_name,
                metadata=asset_node.metadata,
                tags=asset_node.tags,
                group_name=asset_node.group_name,
                freshness_policy=asset_node.freshness_policy,
                is_source=asset_node.is_external,
                is_observable=asset_node.is_observable,
                execution_set_identifier=repo.asset_graph.get_execution_set_identifier(key),
                required_top_level_resources=required_top_level_resources,
                auto_materialize_policy=automation_condition.as_auto_materialize_policy()
                if automation_condition
                else None,
                automation_condition_snapshot=automation_condition_snapshot,
                backfill_policy=asset_node.backfill_policy,
                auto_observe_interval_minutes=asset_node.auto_observe_interval_minutes,
                owners=asset_node.owners,
            )
        )

    return asset_node_snaps


def resource_value_snap_from_raw(v: Any) -> ResourceValueSnap:
    if isinstance(v, dict) and set(v.keys()) == {"env"}:
        return ResourceConfigEnvVarSnap(name=v["env"])
    return json.dumps(v)


def _get_nested_resources_map(
    resource_datas: Mapping[str, ResourceDefinition],
    top_level_resources: Mapping[str, ResourceDefinition],
) -> Mapping[str, Mapping[str, NestedResource]]:
    out_map: Mapping[str, Mapping[str, NestedResource]] = {}
    for resource_name, resource_def in resource_datas.items():
        out_map[resource_name] = _get_nested_resources(resource_def, top_level_resources)
    return out_map


def _find_match(nested_resource, resource_defs) -> Optional[str]:
    if is_coercible_to_resource(nested_resource):
        defn = coerce_to_resource(nested_resource)
    else:
        return None

    for k, v in resource_defs.items():
        if defn is v:
            return k

    return None


def _get_nested_resources(
    resource_def: ResourceDefinition,
    top_level_resources: Mapping[str, ResourceDefinition],
) -> Mapping[str, NestedResource]:
    # ConfigurableResources may have "anonymous" nested resources, which are not
    # explicitly specified as top-level resources
    if isinstance(
        resource_def,
        (
            ConfigurableResourceFactoryResourceDefinition,
            ConfigurableIOManagerFactoryResourceDefinition,
        ),
    ):
        results = {}
        for k, nested_resource in resource_def.nested_resources.items():
            top_level_key = _find_match(nested_resource, top_level_resources)
            if top_level_key:
                results[k] = NestedResource(NestedResourceType.TOP_LEVEL, top_level_key)
            else:
                results[k] = NestedResource(
                    NestedResourceType.ANONYMOUS, nested_resource.__class__.__name__
                )
        return results
    else:
        return {
            k: NestedResource(NestedResourceType.TOP_LEVEL, k)
            for k in resource_def.required_resource_keys
        }


def _get_class_name(cls: type) -> str:
    """Returns the fully qualified class name of the given class."""
    return str(cls)[8:-2]


PARTITION_SET_SNAP_NAME_SUFFIX: Final = "_partition_set"


def partition_set_snap_name_for_job_name(job_name) -> str:
    return f"{job_name}{PARTITION_SET_SNAP_NAME_SUFFIX}"


def job_name_for_partition_set_snap_name(name: str) -> str:
    job_name_len = len(name) - len(PARTITION_SET_SNAP_NAME_SUFFIX)
    return name[:job_name_len]


def active_presets_from_job_def(job_def: JobDefinition) -> Sequence[PresetSnap]:
    check.inst_param(job_def, "job_def", JobDefinition)
    if job_def.run_config is None:
        return []
    else:
        return [
            PresetSnap(
                name=DEFAULT_PRESET_NAME,
                run_config=job_def.run_config,
                op_selection=None,
                mode=DEFAULT_MODE_NAME,
                tags={},
            )
        ]


def resolve_automation_condition_args(
    automation_condition: Optional[AutomationCondition],
) -> tuple[Optional[AutomationCondition], Optional[AutomationConditionSnapshot]]:
    if automation_condition is None:
        return None, None
    elif automation_condition.is_serializable:
        # to avoid serializing too much data, only store the full condition if
        # it is available
        return automation_condition, None
    else:
        # for non-serializable conditions, only include the snapshot
        return None, automation_condition.get_snapshot()


def _extract_fast(serialized_job_data: str):
    target_key = f'"{_JOB_SNAP_STORAGE_FIELD}": '
    target_substr = target_key + get_prefix_for_a_serialized(JobSnap)
    # look for key: type
    idx = serialized_job_data.find(target_substr)
    check.invariant(idx > 0)
    # slice starting after key:
    start_idx = idx + len(target_key)

    # trim outer object }
    # assumption that pipeline_snapshot is last field under test in test_job_data_snap_layout
    serialized_job_snap = serialized_job_data[start_idx:-1]
    check.invariant(serialized_job_snap[0] == "{" and serialized_job_snap[-1] == "}")

    return serialized_job_snap


def _extract_safe(serialized_job_data: str):
    # Intentionally use json directly instead of serdes to avoid losing information if the current process
    # is older than the source process.
    return json.dumps(json.loads(serialized_job_data)[_JOB_SNAP_STORAGE_FIELD])


DISABLE_FAST_EXTRACT_ENV_VAR = "DAGSTER_DISABLE_JOB_SNAP_FAST_EXTRACT"


def extract_serialized_job_snap_from_serialized_job_data_snap(serialized_job_data_snap: str):
    # utility used by DagsterCloudAgent to extract JobSnap out of JobDataSnap
    # efficiently and safely
    if not serialized_job_data_snap.startswith(get_prefix_for_a_serialized(JobDataSnap)):
        raise Exception("Passed in string does not meet expectations for a serialized JobDataSnap")

    if not os.getenv(DISABLE_FAST_EXTRACT_ENV_VAR):
        try:
            return _extract_fast(serialized_job_data_snap)
        except Exception:
            pass

    return _extract_safe(serialized_job_data_snap)



import sys
from collections.abc import Mapping
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.code_pointer import ModuleCodePointer
from dagster._core.definitions.selector import JobSubsetSelector, RepositorySelector
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.origin import RepositoryPythonOrigin
from dagster._core.remote_representation.origin import (
    CodeLocationOrigin,
    RegisteredCodeLocationOrigin,
    RemoteRepositoryOrigin,
)
from dagster._record import record
from dagster._serdes.serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.remote_representation.code_location import CodeLocation


@whitelist_for_serdes
@record
class RepositoryHandle:
    repository_name: str
    code_location_origin: CodeLocationOrigin
    repository_python_origin: RepositoryPythonOrigin
    display_metadata: Mapping[str, str]

    @classmethod
    def from_location(cls, repository_name: str, code_location: "CodeLocation"):
        from dagster._core.remote_representation.code_location import CodeLocation

        check.inst_param(code_location, "code_location", CodeLocation)
        return cls(
            repository_name=repository_name,
            code_location_origin=code_location.origin,
            repository_python_origin=code_location.get_repository_python_origin(repository_name),
            display_metadata=code_location.get_display_metadata(),
        )

    @property
    def location_name(self) -> str:
        return self.code_location_origin.location_name

    def get_remote_origin(self) -> RemoteRepositoryOrigin:
        return RemoteRepositoryOrigin(
            code_location_origin=self.code_location_origin,
            repository_name=self.repository_name,
        )

    def get_python_origin(self) -> RepositoryPythonOrigin:
        return self.repository_python_origin

    def to_selector(self) -> RepositorySelector:
        return RepositorySelector(
            location_name=self.location_name,
            repository_name=self.repository_name,
        )

    def get_compound_id(self) -> "CompoundID":
        return CompoundID(
            remote_origin_id=self.get_remote_origin().get_id(),
            selector_id=self.to_selector().selector_id,
        )

    @staticmethod
    def for_test(
        *,
        location_name: str = "fake_location",
        repository_name: str = "fake_repository",
        display_metadata: Optional[Mapping[str, str]] = None,
    ) -> "RepositoryHandle":
        return RepositoryHandle(
            repository_name=repository_name,
            code_location_origin=RegisteredCodeLocationOrigin(location_name=location_name),
            repository_python_origin=RepositoryPythonOrigin(
                executable_path=sys.executable,
                code_pointer=ModuleCodePointer(
                    module="fake.module",
                    fn_name="fake_fn",
                ),
            ),
            display_metadata=display_metadata or {},
        )


@record(kw_only=False)
class JobHandle:
    job_name: str
    repository_handle: RepositoryHandle

    def to_string(self):
        return f"{self.location_name}.{self.repository_name}.{self.job_name}"

    @property
    def repository_name(self):
        return self.repository_handle.repository_name

    @property
    def location_name(self):
        return self.repository_handle.location_name

    def get_remote_origin(self):
        return self.repository_handle.get_remote_origin().get_job_origin(self.job_name)

    def get_python_origin(self):
        return self.repository_handle.get_python_origin().get_job_origin(self.job_name)

    def to_selector(self) -> JobSubsetSelector:
        return JobSubsetSelector(
            location_name=self.location_name,
            repository_name=self.repository_name,
            job_name=self.job_name,
            op_selection=None,
        )


@record(kw_only=False)
class InstigatorHandle:
    instigator_name: str
    repository_handle: RepositoryHandle

    @property
    def repository_name(self) -> str:
        return self.repository_handle.repository_name

    @property
    def location_name(self) -> str:
        return self.repository_handle.location_name

    def get_remote_origin(self):
        return self.repository_handle.get_remote_origin().get_instigator_origin(
            self.instigator_name
        )


@record
class PartitionSetHandle:
    partition_set_name: str
    repository_handle: RepositoryHandle

    @property
    def repository_name(self):
        return self.repository_handle.repository_name

    @property
    def location_name(self):
        return self.repository_handle.location_name

    def get_remote_origin(self):
        return self.repository_handle.get_remote_origin().get_partition_set_origin(
            self.partition_set_name
        )


_DELIMITER = "::"


@record
class CompoundID:
    """Compound ID object for the two id schemes that state is recorded in the database against."""

    remote_origin_id: str
    selector_id: str

    def to_string(self) -> str:
        return f"{self.remote_origin_id}{_DELIMITER}{self.selector_id}"

    @staticmethod
    def from_string(serialized: str):
        parts = serialized.split(_DELIMITER)
        if len(parts) != 2:
            raise DagsterInvariantViolationError(f"Invalid serialized InstigatorID: {serialized}")

        return CompoundID(
            remote_origin_id=parts[0],
            selector_id=parts[1],
        )

    @staticmethod
    def is_valid_string(serialized: str):
        parts = serialized.split(_DELIMITER)
        return len(parts) == 2



import datetime
import logging
import os
import random
import sys
import threading
from collections import defaultdict
from collections.abc import Generator, Mapping, Sequence
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import AbstractContextManager, ExitStack
from typing import TYPE_CHECKING, NamedTuple, Optional, Union, cast

from typing_extensions import Self

import dagster._check as check
from dagster._core.definitions.run_request import RunRequest
from dagster._core.definitions.schedule_definition import DefaultScheduleStatus
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.definitions.timestamp import TimestampWithTimezone
from dagster._core.errors import DagsterCodeLocationLoadError, DagsterUserCodeUnreachableError
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import RemoteSchedule
from dagster._core.remote_representation.code_location import CodeLocation
from dagster._core.remote_representation.external import RemoteJob
from dagster._core.scheduler.instigation import (
    InstigatorState,
    InstigatorStatus,
    InstigatorTick,
    InstigatorType,
    ScheduleInstigatorData,
    TickData,
    TickStatus,
)
from dagster._core.scheduler.scheduler import DEFAULT_MAX_CATCHUP_RUNS
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunsFilter
from dagster._core.storage.tags import RUN_KEY_TAG, SCHEDULED_EXECUTION_TIME_TAG
from dagster._core.telemetry import SCHEDULED_RUN_CREATED, hash_name, log_action
from dagster._core.utils import InheritContextThreadPoolExecutor
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.utils import DaemonErrorCapture
from dagster._scheduler.stale import resolve_stale_or_missing_assets
from dagster._time import get_current_datetime, get_current_timestamp
from dagster._utils import DebugCrashFlags, SingleInstigatorDebugCrashFlags, check_for_debug_crash
from dagster._utils.error import SerializableErrorInfo
from dagster._utils.log import default_date_format_string
from dagster._utils.merger import merge_dicts

if TYPE_CHECKING:
    from dagster._daemon.daemon import DaemonIterator


# how often do we update the job row in the database with the last iteration timestamp.  This
# creates a checkpoint so that if the cron schedule changes, we don't try to backfill schedule ticks
# from the start of the schedule, just since the last recorded iteration interval.
LAST_ITERATION_CHECKPOINT_INTERVAL_SECONDS = int(
    os.getenv("DAGSTER_SCHEDULE_CHECKPOINT_INTERVAL_SECONDS", "3600")
)

LAST_ITERATION_CHECKPOINT_JITTER_SECONDS = int(
    os.getenv("DAGSTER_SCHEDULE_CHECKPOINT_JITTER_SECONDS", "600")
)

RETAIN_ORPHANED_STATE_INTERVAL_SECONDS = int(
    os.getenv("DAGSTER_SCHEDULE_ORPHANED_STATE_RETENTION_SECONDS", "43200")  # 12 hours
)

# How long to wait if an error is raised in the SchedulerDaemon iteration
ERROR_INTERVAL_TIME = 5


class _ScheduleLaunchContext(AbstractContextManager):
    def __init__(
        self,
        remote_schedule: RemoteSchedule,
        tick: InstigatorTick,
        instance: DagsterInstance,
        logger: logging.Logger,
        tick_retention_settings,
    ):
        self._remote_schedule = remote_schedule
        self._instance = instance
        self._logger = logger
        self._tick = tick
        self._purge_settings = defaultdict(set)
        for status, day_offset in tick_retention_settings.items():
            self._purge_settings[day_offset].add(status)

    @property
    def failure_count(self) -> int:
        return self._tick.tick_data.failure_count

    @property
    def consecutive_failure_count(self) -> int:
        return self._tick.tick_data.consecutive_failure_count or self._tick.tick_data.failure_count

    @property
    def tick_id(self) -> str:
        return str(self._tick.tick_id)

    @property
    def log_key(self) -> Sequence[str]:
        return [
            self._remote_schedule.handle.repository_name,
            self._remote_schedule.name,
            self.tick_id,
        ]

    def update_state(self, status, error=None, **kwargs):
        if status in {TickStatus.SKIPPED, TickStatus.SUCCESS}:
            kwargs["failure_count"] = 0
            kwargs["consecutive_failure_count"] = 0

        skip_reason = kwargs.get("skip_reason")
        if "skip_reason" in kwargs:
            del kwargs["skip_reason"]

        self._tick = self._tick.with_status(status=status, error=error, **kwargs)

        if skip_reason:
            self._tick = self._tick.with_reason(skip_reason=skip_reason)

    def add_run_info(self, run_id=None, run_key=None):
        self._tick = self._tick.with_run_info(run_id, run_key)

    def add_log_key(self, log_key: Sequence[str]) -> None:
        self._tick = self._tick.with_log_key(log_key)

    def _write(self):
        self._instance.update_tick(self._tick)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self._write()
        for day_offset, statuses in self._purge_settings.items():
            if day_offset <= 0:
                continue
            self._instance.purge_ticks(
                self._remote_schedule.get_remote_origin_id(),
                selector_id=self._remote_schedule.selector_id,
                before=(get_current_datetime() - datetime.timedelta(days=day_offset)).timestamp(),
                tick_statuses=list(statuses),
            )


SECONDS_IN_MINUTE = 60


def _get_next_scheduler_iteration_time(start_time: float) -> float:
    # Wait until at least the next minute to run again, since the minimum granularity
    # for a cron schedule is every minute
    last_minute_time = start_time - (start_time % SECONDS_IN_MINUTE)
    return last_minute_time + SECONDS_IN_MINUTE


class ScheduleIterationTimes(NamedTuple):
    """Timestamp information returned by each scheduler iteration that the core scheduler
    loop can use to intelligently schedule the next tick.

    last_iteration_timestamp is used by subsequent evaluations of this schedule to ensure that we
    don't accidentally create incorrect runs after the cronstring changes (it is stored in memory
    in these objects and is also periodically persisted on the schedule row in the database -
    see _write_and_get_next_checkpoint_timestamp).

    next_iteration_timestamp is the timestamp until which the scheduler can wait until running this
    schedule again (assuming the cron schedule has not changed) - either because that's the next
    time we know the schedule needs to run based on the last time we evaluated its cron string,
    or because that's the next time we've determined that we should update the persisted
    last_iteration_timestamp value for this schedule described above (this is written on a fixed
    interval plus a random jitter value to ensure not every schedule tries to do this at once -
    this value is also determined in _write_and_get_next_checkpoint_timestamp.).
    """

    cron_schedule: Union[str, Sequence[str]]
    next_iteration_timestamp: float
    last_iteration_timestamp: float

    def should_run_next_iteration(self, schedule: RemoteSchedule, now_timestamp: float):
        if schedule.cron_schedule != self.cron_schedule:
            # cron schedule has changed - always run next iteration to check
            return True
        return now_timestamp >= self.next_iteration_timestamp


def execute_scheduler_iteration_loop(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    max_catchup_runs: int,
    max_tick_retries: int,
    shutdown_event: threading.Event,
) -> "DaemonIterator":
    from dagster._daemon.daemon import SpanMarker

    scheduler_run_futures: dict[str, Future] = {}
    iteration_times: dict[str, ScheduleIterationTimes] = {}

    threadpool_executor = None
    submit_threadpool_executor = None

    with ExitStack() as stack:
        settings = workspace_process_context.instance.get_scheduler_settings()
        if settings.get("use_threads"):
            threadpool_executor = stack.enter_context(
                InheritContextThreadPoolExecutor(
                    max_workers=settings.get("num_workers"),
                    thread_name_prefix="schedule_daemon_worker",
                )
            )
            num_submit_workers = settings.get("num_submit_workers")
            if num_submit_workers:
                submit_threadpool_executor = stack.enter_context(
                    InheritContextThreadPoolExecutor(
                        max_workers=settings.get("num_submit_workers"),
                        thread_name_prefix="schedule_submit_worker",
                    )
                )

        while True:
            start_time = get_current_timestamp()
            end_datetime_utc = get_current_datetime()

            next_interval_time = _get_next_scheduler_iteration_time(start_time)

            yield SpanMarker.START_SPAN

            try:
                yield from launch_scheduled_runs(
                    workspace_process_context,
                    logger,
                    end_datetime_utc=end_datetime_utc,
                    iteration_times=iteration_times,
                    threadpool_executor=threadpool_executor,
                    submit_threadpool_executor=submit_threadpool_executor,
                    scheduler_run_futures=scheduler_run_futures,
                    max_catchup_runs=max_catchup_runs,
                    max_tick_retries=max_tick_retries,
                )
            except Exception:
                error_info = DaemonErrorCapture.process_exception(
                    exc_info=sys.exc_info(),
                    logger=logger,
                    log_message="SchedulerDaemon caught an error",
                )
                yield error_info
                # Wait a few seconds after an error
                next_interval_time = min(start_time + ERROR_INTERVAL_TIME, next_interval_time)

            yield SpanMarker.END_SPAN

            end_time = get_current_timestamp()
            if next_interval_time > end_time:
                # Sleep until the beginning of the next minute, plus a small epsilon to
                # be sure that we're past the start of the minute
                shutdown_event.wait(next_interval_time - end_time + 0.001)
                yield


def launch_scheduled_runs(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    end_datetime_utc: datetime.datetime,
    iteration_times: dict[str, ScheduleIterationTimes],
    threadpool_executor: Optional[ThreadPoolExecutor] = None,
    submit_threadpool_executor: Optional[ThreadPoolExecutor] = None,
    scheduler_run_futures: Optional[dict[str, Future]] = None,
    max_catchup_runs: int = DEFAULT_MAX_CATCHUP_RUNS,
    max_tick_retries: int = 0,
    debug_crash_flags: Optional[DebugCrashFlags] = None,
) -> "DaemonIterator":
    instance = workspace_process_context.instance

    current_workspace = {
        location_entry.origin.location_name: location_entry
        for location_entry in workspace_process_context.create_request_context()
        .get_code_location_entries()
        .values()
    }

    all_schedule_states = {
        schedule_state.selector_id: schedule_state
        for schedule_state in instance.all_instigator_state(instigator_type=InstigatorType.SCHEDULE)
    }

    tick_retention_settings = instance.get_tick_retention_settings(InstigatorType.SCHEDULE)

    running_schedules: dict[str, RemoteSchedule] = {}
    all_workspace_schedule_selector_ids = set()
    error_locations = set()

    now_timestamp = end_datetime_utc.timestamp()

    for location_entry in current_workspace.values():
        code_location = location_entry.code_location
        if code_location:
            for repo in code_location.get_repositories().values():
                for schedule in repo.get_schedules():
                    selector_id = schedule.selector_id
                    all_workspace_schedule_selector_ids.add(selector_id)
                    if schedule.get_current_instigator_state(
                        all_schedule_states.get(selector_id)
                    ).is_running:
                        running_schedules[selector_id] = schedule
                    elif all_schedule_states.get(selector_id):
                        schedule_state = all_schedule_states[selector_id]
                        # If there is a DB row to update, see if we should still update the
                        # last_iteration_timestamp
                        _write_and_get_next_checkpoint_timestamp(
                            instance,
                            all_schedule_states[selector_id],
                            cast(ScheduleInstigatorData, schedule_state.instigator_data),
                            now_timestamp,
                        )

        elif location_entry.load_error:
            error_locations.add(location_entry.origin.location_name)

    # Remove any schedule states that were previously created and can no longer
    # be found in the workspace (so that if they are later added back again,
    # their timestamps will start at the correct place)
    states_to_delete = [
        schedule_state
        for selector_id, schedule_state in all_schedule_states.items()
        if selector_id not in all_workspace_schedule_selector_ids
        or (
            schedule_state.status == InstigatorStatus.DECLARED_IN_CODE
            and not running_schedules.get(selector_id)
        )
    ]
    for state in states_to_delete:
        location_name = state.origin.repository_origin.code_location_origin.location_name

        if location_name in error_locations:
            # don't clean up state if its location is an error state
            continue

        _last_iteration_time = (
            state.instigator_data.last_iteration_timestamp or 0.0
            if isinstance(state.instigator_data, ScheduleInstigatorData)
            else 0.0
        )

        # Remove all-stopped states declared in code immediately.
        # Also remove all other states that are not present in the workspace after a 12-hour grace period.
        if state.status == InstigatorStatus.DECLARED_IN_CODE or (
            _last_iteration_time
            and _last_iteration_time + RETAIN_ORPHANED_STATE_INTERVAL_SECONDS
            < end_datetime_utc.timestamp()
        ):
            logger.info(
                f"Removing state for schedule {state.instigator_name} that is "
                f"no longer present in {location_name}."
            )
            instance.delete_instigator_state(state.instigator_origin_id, state.selector_id)

    if not running_schedules:
        yield
        return

    for schedule in running_schedules.values():
        error_info = None
        try:
            schedule_state = all_schedule_states.get(schedule.selector_id)
            if not schedule_state:
                assert schedule.default_status == DefaultScheduleStatus.RUNNING
                schedule_state = InstigatorState(
                    schedule.get_remote_origin(),
                    InstigatorType.SCHEDULE,
                    InstigatorStatus.DECLARED_IN_CODE,
                    ScheduleInstigatorData(
                        schedule.cron_schedule,
                        end_datetime_utc.timestamp(),
                    ),
                )
                instance.add_instigator_state(schedule_state)

            schedule_debug_crash_flags = (
                debug_crash_flags.get(schedule_state.instigator_name) if debug_crash_flags else None
            )

            if threadpool_executor:
                if scheduler_run_futures is None:
                    check.failed(
                        "scheduler_run_futures dict must be passed with threadpool_executor"
                    )

                if schedule.selector_id in scheduler_run_futures:
                    if scheduler_run_futures[schedule.selector_id].done():
                        try:
                            result = scheduler_run_futures[schedule.selector_id].result()
                            iteration_times[schedule.selector_id] = result
                        except Exception:
                            # Log exception and continue on rather than erroring the whole scheduler loop

                            DaemonErrorCapture.process_exception(
                                exc_info=sys.exc_info(),
                                logger=logger,
                                log_message=f"Error getting tick result for schedule {schedule.name}",
                            )
                        del scheduler_run_futures[schedule.selector_id]
                    else:
                        # only allow one tick per schedule to be in flight
                        continue

                previous_iteration_times = iteration_times.get(schedule.selector_id)
                if (
                    previous_iteration_times
                    and not previous_iteration_times.should_run_next_iteration(
                        schedule, end_datetime_utc.timestamp()
                    )
                ):
                    # Not enough time has passed for this schedule, don't bother creating a thread
                    continue

                future = threadpool_executor.submit(
                    launch_scheduled_runs_for_schedule,
                    workspace_process_context,
                    logger,
                    schedule,
                    schedule_state,
                    end_datetime_utc,
                    max_catchup_runs,
                    max_tick_retries,
                    tick_retention_settings,
                    schedule_debug_crash_flags,
                    submit_threadpool_executor=submit_threadpool_executor,
                    in_memory_last_iteration_timestamp=(
                        previous_iteration_times.last_iteration_timestamp
                        if previous_iteration_times
                        else None
                    ),
                )
                scheduler_run_futures[schedule.selector_id] = future
                yield

            else:
                previous_iteration_times = iteration_times.get(schedule.selector_id)
                if (
                    previous_iteration_times
                    and not previous_iteration_times.should_run_next_iteration(
                        schedule, end_datetime_utc.timestamp()
                    )
                ):
                    # Not enough time has passed for this schedule, don't bother executing
                    continue

                # evaluate the schedules in a loop, synchronously, yielding to allow the schedule daemon to
                # heartbeat
                found_iteration_times = False
                for yielded_value in launch_scheduled_runs_for_schedule_iterator(
                    workspace_process_context,
                    logger,
                    schedule,
                    schedule_state,
                    end_datetime_utc,
                    max_catchup_runs,
                    max_tick_retries,
                    tick_retention_settings,
                    schedule_debug_crash_flags,
                    submit_threadpool_executor=None,
                    in_memory_last_iteration_timestamp=(
                        previous_iteration_times.last_iteration_timestamp
                        if previous_iteration_times
                        else None
                    ),
                ):
                    if isinstance(yielded_value, ScheduleIterationTimes):
                        check.invariant(
                            not found_iteration_times,
                            "launch_scheduled_runs_for_schedule_iterator yielded more than one ScheduleIterationTimes",
                        )
                        found_iteration_times = True
                        iteration_times[schedule.selector_id] = yielded_value
                    else:
                        yield yielded_value
                check.invariant(
                    found_iteration_times,
                    "launch_scheduled_runs_for_schedule_iterator did not yield a ScheduleIterationTimes",
                )
        except Exception:
            error_info = DaemonErrorCapture.process_exception(
                exc_info=sys.exc_info(),
                logger=logger,
                log_message=f"Scheduler caught an error for schedule {schedule.name}",
            )

        yield error_info


def launch_scheduled_runs_for_schedule(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    remote_schedule: RemoteSchedule,
    schedule_state: InstigatorState,
    end_datetime_utc: datetime.datetime,
    max_catchup_runs: int,
    max_tick_retries: int,
    tick_retention_settings: Mapping[TickStatus, int],
    schedule_debug_crash_flags: Optional[SingleInstigatorDebugCrashFlags],
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
    in_memory_last_iteration_timestamp: Optional[float],
) -> ScheduleIterationTimes:
    # evaluate the tick immediately, but from within a thread.  The main thread should be able to
    # heartbeat to keep the daemon alive
    iteration_times = None
    for yielded_value in launch_scheduled_runs_for_schedule_iterator(
        workspace_process_context,
        logger,
        remote_schedule,
        schedule_state,
        end_datetime_utc,
        max_catchup_runs,
        max_tick_retries,
        tick_retention_settings,
        schedule_debug_crash_flags,
        submit_threadpool_executor=submit_threadpool_executor,
        in_memory_last_iteration_timestamp=in_memory_last_iteration_timestamp,
    ):
        if isinstance(yielded_value, ScheduleIterationTimes):
            iteration_times = yielded_value

    return check.not_none(iteration_times)


def launch_scheduled_runs_for_schedule_iterator(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    remote_schedule: RemoteSchedule,
    schedule_state: InstigatorState,
    end_datetime_utc: datetime.datetime,
    max_catchup_runs: int,
    max_tick_retries: int,
    tick_retention_settings: Mapping[TickStatus, int],
    schedule_debug_crash_flags: Optional[SingleInstigatorDebugCrashFlags],
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
    in_memory_last_iteration_timestamp: Optional[float],
) -> Generator[Union[None, SerializableErrorInfo, ScheduleIterationTimes], None, None]:
    schedule_state = check.inst_param(schedule_state, "schedule_state", InstigatorState)
    end_datetime_utc = check.inst_param(end_datetime_utc, "end_datetime_utc", datetime.datetime)
    instance = workspace_process_context.instance

    instigator_origin_id = remote_schedule.get_remote_origin_id()
    ticks = instance.get_ticks(instigator_origin_id, remote_schedule.selector_id, limit=1)
    latest_tick: Optional[InstigatorTick] = ticks[0] if ticks else None

    instigator_data = cast(ScheduleInstigatorData, schedule_state.instigator_data)
    start_timestamp_utc: float = instigator_data.start_timestamp or 0

    if latest_tick:
        if latest_tick.status == TickStatus.STARTED or (
            latest_tick.status == TickStatus.FAILURE
            and latest_tick.failure_count <= max_tick_retries
        ):
            # Scheduler was interrupted while performing this tick, re-do it
            start_timestamp_utc = max(
                start_timestamp_utc,
                latest_tick.timestamp,
                instigator_data.last_iteration_timestamp or 0.0,
                in_memory_last_iteration_timestamp or 0.0,
            )
        else:
            start_timestamp_utc = max(
                start_timestamp_utc,
                latest_tick.timestamp + 1,
                instigator_data.last_iteration_timestamp or 0.0,
                in_memory_last_iteration_timestamp or 0.0,
            )
    else:
        start_timestamp_utc = max(
            start_timestamp_utc,
            instigator_data.last_iteration_timestamp or 0.0,
            in_memory_last_iteration_timestamp or 0.0,
        )

    schedule_name = remote_schedule.name

    timezone_str = remote_schedule.execution_timezone
    if not timezone_str:
        timezone_str = "UTC"

    tick_times: list[datetime.datetime] = []

    now_timestamp = end_datetime_utc.timestamp()

    next_iteration_timestamp = None

    for next_time in remote_schedule.execution_time_iterator(start_timestamp_utc):
        next_tick_timestamp = next_time.timestamp()
        if next_tick_timestamp > now_timestamp:
            next_iteration_timestamp = next_tick_timestamp
            break

        tick_times.append(next_time)

    if not tick_times:
        next_checkpoint_timestamp = _write_and_get_next_checkpoint_timestamp(
            instance,
            schedule_state,
            instigator_data,
            now_timestamp,
        )

        next_iteration_timestamp = min(
            check.not_none(next_iteration_timestamp), next_checkpoint_timestamp
        )

        yield ScheduleIterationTimes(
            cron_schedule=remote_schedule.cron_schedule,
            next_iteration_timestamp=next_iteration_timestamp,
            last_iteration_timestamp=now_timestamp,
        )
        return

    if not remote_schedule.partition_set_name and len(tick_times) > 1:
        logger.warning(f"{schedule_name} has no partition set, so not trying to catch up")
        tick_times = tick_times[-1:]
    elif len(tick_times) > max_catchup_runs:
        logger.warning(f"{schedule_name} has fallen behind, only launching {max_catchup_runs} runs")
        tick_times = tick_times[-max_catchup_runs:]

    if len(tick_times) == 1:
        tick_time = tick_times[0].strftime(default_date_format_string())
        logger.info(f"Evaluating schedule `{schedule_name}` at {tick_time}")
    else:
        times = ", ".join([time.strftime(default_date_format_string()) for time in tick_times])
        logger.info(f"Evaluating schedule `{schedule_name}` at the following times: {times}")

    for schedule_time in tick_times:
        schedule_timestamp = schedule_time.timestamp()
        schedule_time_str = schedule_time.strftime(default_date_format_string())

        consecutive_failure_count = 0
        if latest_tick and latest_tick.status in {TickStatus.FAILURE, TickStatus.STARTED}:
            consecutive_failure_count = (
                latest_tick.consecutive_failure_count or latest_tick.failure_count
            )

        if latest_tick and latest_tick.timestamp == schedule_timestamp:
            tick = latest_tick
            if latest_tick.status == TickStatus.FAILURE:
                logger.info(f"Retrying previously failed schedule execution at {schedule_time_str}")
            else:
                logger.info(
                    f"Resuming previously interrupted schedule execution at {schedule_time_str}"
                )
        else:
            tick = instance.create_tick(
                TickData(
                    instigator_origin_id=instigator_origin_id,
                    instigator_name=schedule_name,
                    instigator_type=InstigatorType.SCHEDULE,
                    status=TickStatus.STARTED,
                    timestamp=schedule_timestamp,
                    selector_id=remote_schedule.selector_id,
                    consecutive_failure_count=consecutive_failure_count,
                )
            )

            check_for_debug_crash(schedule_debug_crash_flags, "TICK_CREATED")

        with _ScheduleLaunchContext(
            remote_schedule, tick, instance, logger, tick_retention_settings
        ) as tick_context:
            try:
                check_for_debug_crash(schedule_debug_crash_flags, "TICK_HELD")
                tick_context.add_log_key(tick_context.log_key)

                yield from _schedule_runs_at_time(
                    workspace_process_context,
                    logger,
                    remote_schedule,
                    schedule_time,
                    timezone_str,
                    tick_context,
                    submit_threadpool_executor,
                    schedule_debug_crash_flags,
                )
            except Exception as e:
                if isinstance(e, (DagsterUserCodeUnreachableError, DagsterCodeLocationLoadError)):
                    try:
                        raise DagsterUserCodeUnreachableError(
                            f"Unable to reach the user code server for schedule {schedule_name}."
                            " Schedule will resume execution once the server is available."
                        ) from e
                    except:
                        error_data = DaemonErrorCapture.process_exception(
                            sys.exc_info(),
                            logger=logger,
                            log_message=f"Scheduler daemon caught an error for schedule {remote_schedule.name}",
                        )
                        tick_context.update_state(
                            TickStatus.FAILURE,
                            error=error_data,
                            # don't increment the failure count - retry forever until the server comes back up
                            # or the schedule is turned off
                            failure_count=tick_context.failure_count,
                            consecutive_failure_count=tick_context.consecutive_failure_count + 1,
                        )
                        yield error_data
                else:
                    error_data = DaemonErrorCapture.process_exception(
                        sys.exc_info(),
                        logger=logger,
                        log_message=f"Scheduler daemon caught an error for schedule {remote_schedule.name}",
                    )
                    tick_context.update_state(
                        TickStatus.FAILURE,
                        error=error_data,
                        failure_count=tick_context.failure_count + 1,
                        consecutive_failure_count=tick_context.consecutive_failure_count + 1,
                    )
                    yield error_data

                # Plan to run the same tick again using the schedule timestamp
                # as both the next_iteration_timestamp and the last_iteration_timestmap
                # (to ensure that the scheduler doesn't accidentally skip past it)
                yield ScheduleIterationTimes(
                    cron_schedule=remote_schedule.cron_schedule,
                    next_iteration_timestamp=schedule_time.timestamp(),
                    last_iteration_timestamp=schedule_time.timestamp(),
                )
                return

    # now log the iteration timestamp
    next_checkpoint_timestamp = _write_and_get_next_checkpoint_timestamp(
        instance,
        schedule_state,
        instigator_data,
        end_datetime_utc.timestamp(),
    )
    next_iteration_timestamp = min(
        check.not_none(next_iteration_timestamp), next_checkpoint_timestamp
    )
    yield ScheduleIterationTimes(
        cron_schedule=remote_schedule.cron_schedule,
        next_iteration_timestamp=next_iteration_timestamp,
        last_iteration_timestamp=now_timestamp,
    )
    return


class SubmitRunRequestResult(NamedTuple):
    run_key: Optional[str]
    error_info: Optional[SerializableErrorInfo]
    existing_run: Optional[DagsterRun]
    submitted_run: Optional[DagsterRun]


def _submit_run_request(
    run_request: RunRequest,
    workspace_process_context: IWorkspaceProcessContext,
    remote_schedule: RemoteSchedule,
    schedule_time: datetime.datetime,
    logger,
    debug_crash_flags,
) -> SubmitRunRequestResult:
    instance = workspace_process_context.instance
    schedule_origin = remote_schedule.get_remote_origin()

    run = _get_existing_run_for_request(instance, remote_schedule, schedule_time, run_request)
    if run:
        if run.status != DagsterRunStatus.NOT_STARTED:
            # A run already exists and was launched for this time period,
            # but the scheduler must have crashed or errored before the tick could be put
            # into a SUCCESS state
            logger.info(
                f"Run {run.run_id} already completed for this execution of {remote_schedule.name}"
            )
            return SubmitRunRequestResult(
                run_key=run_request.run_key, error_info=None, existing_run=run, submitted_run=None
            )
        else:
            logger.info(
                f"Run {run.run_id} already created for this execution of {remote_schedule.name}"
            )
    else:
        job_subset_selector = JobSubsetSelector(
            location_name=schedule_origin.repository_origin.code_location_origin.location_name,
            repository_name=schedule_origin.repository_origin.repository_name,
            job_name=remote_schedule.job_name,
            op_selection=remote_schedule.op_selection,
            asset_selection=run_request.asset_selection,
        )

        # reload the code_location on each submission, request_context derived data can become out date
        # * non-threaded: if number of serial submissions is too many
        # * threaded: if thread sits pending in pool too long
        code_location = _get_code_location_for_schedule(workspace_process_context, remote_schedule)

        remote_job = code_location.get_job(job_subset_selector)

        run = _create_scheduler_run(
            instance,
            schedule_time,
            code_location,
            remote_schedule,
            remote_job,
            run_request,
        )

    check_for_debug_crash(debug_crash_flags, "RUN_CREATED")

    error_info = None

    if run.status != DagsterRunStatus.FAILURE:
        try:
            instance.submit_run(run.run_id, workspace_process_context.create_request_context())
            logger.info(
                f"Completed scheduled launch of run {run.run_id} for {remote_schedule.name}"
            )
        except Exception:
            error_info = DaemonErrorCapture.process_exception(
                exc_info=sys.exc_info(),
                logger=logger,
                log_message=f"Run {run.run_id} created successfully but failed to launch",
            )

    return SubmitRunRequestResult(
        run_key=run_request.run_key,
        error_info=error_info,
        existing_run=None,
        submitted_run=run,
    )


def _get_code_location_for_schedule(
    workspace_process_context: IWorkspaceProcessContext,
    remote_schedule: RemoteSchedule,
) -> CodeLocation:
    schedule_origin = remote_schedule.get_remote_origin()
    return workspace_process_context.create_request_context().get_code_location(
        schedule_origin.repository_origin.code_location_origin.location_name
    )


def _schedule_runs_at_time(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    remote_schedule: RemoteSchedule,
    schedule_time: datetime.datetime,
    timezone_str: str,
    tick_context: _ScheduleLaunchContext,
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
    debug_crash_flags: Optional[SingleInstigatorDebugCrashFlags] = None,
) -> Generator[Union[None, SerializableErrorInfo, ScheduleIterationTimes], None, None]:
    instance = workspace_process_context.instance
    repository_handle = remote_schedule.handle.repository_handle

    code_location = _get_code_location_for_schedule(workspace_process_context, remote_schedule)

    schedule_execution_data = code_location.get_schedule_execution_data(
        instance=instance,
        repository_handle=repository_handle,
        schedule_name=remote_schedule.name,
        scheduled_execution_time=TimestampWithTimezone(
            schedule_time.timestamp(),
            timezone_str,
        ),
        log_key=tick_context.log_key,
    )
    yield None

    # Kept for backwards compatibility with schedule log keys that were previously created in the
    # schedule evaluation, rather than upfront.
    #
    # Note that to get schedule logs for failed schedule evaluations, we force users to update their
    # Dagster version.
    if schedule_execution_data.log_key:
        tick_context.add_log_key(schedule_execution_data.log_key)

    if not schedule_execution_data.run_requests:
        if schedule_execution_data.skip_message:
            logger.info(
                f"Schedule {remote_schedule.name} skipped: {schedule_execution_data.skip_message}"
            )
        else:
            logger.info(f"No run requests returned for {remote_schedule.name}, skipping")

        # Update tick to skipped state and return
        tick_context.update_state(
            TickStatus.SKIPPED, skip_reason=schedule_execution_data.skip_message
        )
        return

    run_requests = []

    for raw_run_request in schedule_execution_data.run_requests:
        run_request = raw_run_request.with_replaced_attrs(
            tags=merge_dicts(
                raw_run_request.tags,
                DagsterRun.tags_for_tick_id(tick_context.tick_id),
            )
        )

        if run_request.stale_assets_only:
            stale_assets = resolve_stale_or_missing_assets(
                workspace_process_context,  # type: ignore
                run_request,
                remote_schedule,
            )
            # asset selection is empty set after filtering for stale
            if len(stale_assets) == 0:
                continue
            else:
                run_request = run_request.with_replaced_attrs(
                    asset_selection=stale_assets, stale_assets_only=False
                )

        run_requests.append(run_request)

    submit_run_request = lambda run_request: _submit_run_request(
        run_request,
        workspace_process_context,
        remote_schedule,
        schedule_time,
        logger,
        debug_crash_flags,
    )

    if submit_threadpool_executor:
        gen_run_request_results = submit_threadpool_executor.map(submit_run_request, run_requests)
    else:
        gen_run_request_results = map(submit_run_request, run_requests)

    for run_request_result in gen_run_request_results:
        yield run_request_result.error_info

        if run_request_result.existing_run:
            tick_context.add_run_info(
                run_id=run_request_result.existing_run.run_id, run_key=run_request_result.run_key
            )
        else:
            run = check.not_none(run_request_result.submitted_run)
            check_for_debug_crash(debug_crash_flags, "RUN_LAUNCHED")
            tick_context.add_run_info(run_id=run.run_id, run_key=run_request_result.run_key)
            check_for_debug_crash(debug_crash_flags, "RUN_ADDED")

    check_for_debug_crash(debug_crash_flags, "TICK_SUCCESS")
    tick_context.update_state(TickStatus.SUCCESS)


def _get_existing_run_for_request(
    instance: DagsterInstance,
    remote_schedule: RemoteSchedule,
    schedule_time: datetime.datetime,
    run_request: RunRequest,
) -> Optional[DagsterRun]:
    tags = merge_dicts(
        DagsterRun.tags_for_schedule(remote_schedule),
        {
            SCHEDULED_EXECUTION_TIME_TAG: schedule_time.astimezone(
                datetime.timezone.utc
            ).isoformat(),
        },
    )
    if run_request.run_key:
        tags[RUN_KEY_TAG] = run_request.run_key
    runs_filter = RunsFilter(tags=tags)
    existing_runs = instance.get_runs(runs_filter)

    # filter down to match schedule namespace (repository)
    matching_runs = []
    for run in existing_runs:
        # if the run doesn't have an origin consider it a match
        if run.remote_job_origin is None:
            matching_runs.append(run)
        # otherwise prevent the same named schedule (with the same execution time) across repos from effecting each other
        elif (
            remote_schedule.get_remote_origin().repository_origin.get_selector_id()
            == run.remote_job_origin.repository_origin.get_selector_id()
        ):
            matching_runs.append(run)

    if not len(matching_runs):
        return None

    return matching_runs[0]


def _create_scheduler_run(
    instance: DagsterInstance,
    schedule_time: datetime.datetime,
    code_location: CodeLocation,
    remote_schedule: RemoteSchedule,
    remote_job: RemoteJob,
    run_request: RunRequest,
) -> DagsterRun:
    from dagster._daemon.daemon import get_telemetry_daemon_session_id

    run_config = run_request.run_config
    schedule_tags = run_request.tags

    remote_execution_plan = code_location.get_execution_plan(
        remote_job,
        run_config,
        step_keys_to_execute=None,
        known_state=None,
    )
    execution_plan_snapshot = remote_execution_plan.execution_plan_snapshot

    tags = {
        **remote_job.run_tags,
        **schedule_tags,
    }

    tags[SCHEDULED_EXECUTION_TIME_TAG] = schedule_time.astimezone(datetime.timezone.utc).isoformat()
    if run_request.run_key:
        tags[RUN_KEY_TAG] = run_request.run_key

    log_action(
        instance,
        SCHEDULED_RUN_CREATED,
        metadata={
            "DAEMON_SESSION_ID": get_telemetry_daemon_session_id(),
            "SCHEDULE_NAME_HASH": hash_name(remote_schedule.name),
            "repo_hash": hash_name(code_location.name),
            "pipeline_name_hash": hash_name(remote_job.name),
        },
    )

    return instance.create_run(
        job_name=remote_schedule.job_name,
        run_id=None,
        run_config=run_config,
        resolved_op_selection=remote_job.resolved_op_selection,
        step_keys_to_execute=None,
        op_selection=remote_job.op_selection,
        status=DagsterRunStatus.NOT_STARTED,
        root_run_id=None,
        parent_run_id=None,
        tags=tags,
        job_snapshot=remote_job.job_snapshot,
        execution_plan_snapshot=execution_plan_snapshot,
        parent_job_snapshot=remote_job.parent_job_snapshot,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
        asset_selection=(
            frozenset(run_request.asset_selection) if run_request.asset_selection else None
        ),
        asset_check_selection=None,
        asset_graph=code_location.get_repository(
            remote_job.repository_handle.repository_name
        ).asset_graph,
    )


def _write_and_get_next_checkpoint_timestamp(
    instance: DagsterInstance,
    schedule_state: InstigatorState,
    instigator_data: ScheduleInstigatorData,
    iteration_timestamp: float,
) -> float:
    # Utility function that writes iteration timestamps for schedules, to record a
    # successful iteration, regardless of whether or not a tick was processed or not.  This is so
    # that when a cron schedule changes or a schedule changes state, we can modify the evaluation
    # "start time" from the moment that the schedule was turned on to the last time that the
    # schedule was processed in a valid state (even in between ticks).

    # Rather than logging every single iteration, we log every hour.  This means that if the cron
    # schedule changes to run to a time that is less than an hour ago, when the code location is
    # deployed, a tick might be registered for that time, with a run kicking off.

    # Returns the next timestamp that we should plan to log the last_iteration_timestamp - with some
    # additional jitter so that threads won't all come back at the exact same time
    random_jitter_offset = random.randint(0, LAST_ITERATION_CHECKPOINT_JITTER_SECONDS)

    if (
        not instigator_data.last_iteration_timestamp
        or instigator_data.last_iteration_timestamp + LAST_ITERATION_CHECKPOINT_INTERVAL_SECONDS
        <= iteration_timestamp
    ):
        instance.update_instigator_state(
            schedule_state.with_data(
                ScheduleInstigatorData(
                    cron_schedule=instigator_data.cron_schedule,
                    start_timestamp=instigator_data.start_timestamp,
                    last_iteration_timestamp=iteration_timestamp,
                )
            )
        )
        return (
            iteration_timestamp + LAST_ITERATION_CHECKPOINT_INTERVAL_SECONDS + random_jitter_offset
        )

    return (
        instigator_data.last_iteration_timestamp
        + LAST_ITERATION_CHECKPOINT_INTERVAL_SECONDS
        + random_jitter_offset
    )



import collections.abc
import sys
from collections.abc import Iterable, Mapping
from contextlib import contextmanager
from contextvars import ContextVar
from typing import (
    Annotated,
    Any,
    Callable,
    ForwardRef,
    Generic,
    Literal,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from dagster._check.functions import CheckError, TypeOrTupleOfTypes, failed, invariant

try:
    # this type only exists in python 3.10+
    from types import UnionType  # type: ignore
except ImportError:
    UnionType = Union

NoneType = type(None)

_contextual_ns: ContextVar[Mapping[str, type]] = ContextVar("_contextual_ns", default={})
INJECTED_DEFAULT_VALS_LOCAL_VAR = "__injected_defaults__"


class ImportFrom(NamedTuple):
    """A pointer to where to lazily import from to resolve a ForwardRef.

    Used with Annotated ie: Annotated['Foo', ImportFrom('baz.bar')]
    """

    module: str


class _LazyImportPlaceholder: ...


class EvalContext(NamedTuple):
    """Utility class for managing references to global and local namespaces.

    These namespaces are passed to ForwardRef._evaluate to resolve the actual
    type from a string.
    """

    global_ns: dict
    local_ns: dict
    lazy_imports: dict

    @staticmethod
    def capture_from_frame(
        depth: int,
        *,
        add_to_local_ns: Optional[Mapping[str, Any]] = None,
    ) -> "EvalContext":
        """Capture the global and local namespaces via the stack frame.

        Args:
            depth: which stack frame to reference, with depth 0 being the callsite.
            add_to_local_ns: A mapping of additional values to update the local namespace with.

        """
        ctx_frame = sys._getframe(depth + 1)  # noqa # surprisingly not costly

        # copy to not mess up frame data
        global_ns = ctx_frame.f_globals.copy()
        local_ns = ctx_frame.f_locals.copy()

        if add_to_local_ns:
            local_ns.update(add_to_local_ns)

        return EvalContext(
            global_ns=global_ns,
            local_ns=local_ns,
            lazy_imports={},
        )

    @staticmethod
    @contextmanager
    def contextual_namespace(ns: Mapping[str, type]):
        token = _contextual_ns.set(ns)
        try:
            yield
        finally:
            _contextual_ns.reset(token)

    def update_from_frame(self, depth: int):
        # Update the global and local namespaces with symbols from the target frame
        ctx_frame = sys._getframe(depth + 1)  # noqa # surprisingly not costly
        self.global_ns.update(ctx_frame.f_globals)
        self.local_ns.update(ctx_frame.f_locals)

    def register_lazy_import(self, type_name: str, module: str):
        invariant(
            self.lazy_imports.get(type_name, module) == module,
            f"Conflict in lazy imports for type {type_name}, tried to overwrite "
            f"{self.lazy_imports.get(type_name)} with {module}.",
        )
        self.lazy_imports[type_name] = module

    def get_merged_ns(self):
        return {
            **_contextual_ns.get(),
            **self.global_ns,
            **self.local_ns,
        }

    def eval_forward_ref(self, ref: ForwardRef) -> Optional[type]:
        if ref.__forward_arg__ in self.lazy_imports:
            # if we are going to add a lazy import for the type,
            # return a placeholder to grab the name from
            return type(ref.__forward_arg__, (_LazyImportPlaceholder,), {})
        try:
            if sys.version_info <= (3, 9):
                return ref._evaluate(  # noqa # type: ignore
                    globalns=self.get_merged_ns(),
                    localns={},
                )
            else:
                return ref._evaluate(  # noqa
                    globalns=self.get_merged_ns(),
                    localns={},
                    recursive_guard=frozenset(),
                )
        except NameError as e:
            raise CheckError(
                f"Unable to resolve {ref}, could not map string name to actual type using captured frames. "
                f"Use Annotated['{ref.__forward_arg__}', ImportFrom('module.to.import.from')] to create a lazy import."
            ) from e

    def compile_fn(self, body: str, fn_name: str) -> Callable:
        local_ns = {}
        exec(
            body,
            self.get_merged_ns(),
            local_ns,
        )
        return local_ns[fn_name]


T = TypeVar("T")


class _GenClass(Generic[T]): ...


# use a sample to avoid direct private imports (_GenericAlias)
_SampleGeneric = _GenClass[str]


def _coerce_type(
    ttype: Optional[TypeOrTupleOfTypes],
    eval_ctx: EvalContext,
) -> Optional[TypeOrTupleOfTypes]:
    # coerce input type in to the type we want to pass to the check call

    # Any type translates to passing None for the of_type argument
    if ttype is Any or ttype is None:
        return None

    # assume naked strings should be ForwardRefs
    if isinstance(ttype, str):
        return eval_ctx.eval_forward_ref(ForwardRef(ttype))

    if isinstance(ttype, ForwardRef):
        return eval_ctx.eval_forward_ref(ttype)

    if isinstance(ttype, TypeVar):
        return _coerce_type(ttype.__bound__, eval_ctx) if ttype.__bound__ else None

    origin = get_origin(ttype)
    args = get_args(ttype)

    if _is_annotated(origin, args):
        _process_annotated(ttype, args, eval_ctx)
        return _coerce_type(args[0], eval_ctx)

    # cant do isinstance against TypeDict (and we cant subclass check for it)
    # so just coerce any dict subclasses in to dict
    if isinstance(ttype, type) and issubclass(ttype, dict):
        return dict

    # Unions should become a tuple of types to pass to the of_type argument
    # ultimately used as second arg in isinstance(target, tuple_of_types)
    if origin in (UnionType, Union):
        union_types = get_args(ttype)
        coerced_types = []
        for t in union_types:
            # coerce all the inner types
            coerced = _coerce_type(t, eval_ctx)
            if coerced is None or isinstance(coerced, tuple):
                failed(f"Unable to coerce Union member {t} in {ttype}")
            coerced_types.append(coerced)

        return tuple(t for t in coerced_types)

    return ttype


def _container_pair_args(
    args: tuple[type, ...], eval_ctx
) -> tuple[Optional[TypeOrTupleOfTypes], Optional[TypeOrTupleOfTypes]]:
    # process tuple of types as if its two arguments to a container type

    if len(args) == 2:
        return _coerce_type(args[0], eval_ctx), _coerce_type(args[1], eval_ctx)

    return None, None


def _container_single_arg(
    args: tuple[type, ...], eval_ctx: EvalContext
) -> Optional[TypeOrTupleOfTypes]:
    # process tuple of types as if its the single argument to a container type

    if len(args) == 1:
        return _coerce_type(args[0], eval_ctx)

    return None


def _name(target: Optional[TypeOrTupleOfTypes]) -> str:
    # turn a type or tuple of types in to its string representation for printing

    if target is None:
        return "None"

    if target is NoneType:
        return "check.NoneType"

    if isinstance(target, tuple):
        return f"({', '.join(_name(tup_type) for tup_type in target)})"

    if hasattr(target, "__name__"):
        return target.__name__

    if hasattr(target, "_name"):
        n = getattr(target, "_name")
        if n is not None:
            return n

    # If a generic falls through to here, just target the base class
    # and ignore the type hint (for now).
    # Use a sample generic to avoid custom py version handling
    if target.__class__ is _SampleGeneric.__class__:
        return _name(get_origin(target))

    failed(f"Could not calculate string name for {target}")


def _is_annotated(origin, args):
    # 3.9+: origin is Annotated, 3.8: origin == args[0]
    return (origin is Annotated and args) or (len(args) == 1 and args[0] == origin)


def _process_annotated(ttype, args, eval_ctx: EvalContext):
    target_type = args[0]
    # 3.9+: args[1:] has Annotated args, 3.8: its in __metadata__
    annotated_args = getattr(ttype, "__metadata__", args[1:])
    for arg in annotated_args:
        if isinstance(arg, ImportFrom):
            if isinstance(target_type, ForwardRef):
                eval_ctx.register_lazy_import(args[0].__forward_arg__, arg.module)
            elif isinstance(target_type, str):
                eval_ctx.register_lazy_import(args[0], arg.module)
            else:
                failed(
                    f"ImportFrom in Annotated expected to be used with string or ForwardRef only, got {args[0]}",
                )


def build_check_call_str(
    ttype: type,
    name: str,
    eval_ctx: EvalContext,
) -> str:
    from dagster._record import is_record

    # assumes this module is in global/local scope as check
    origin = get_origin(ttype)
    args = get_args(ttype)

    # scalars
    if origin is None:
        if ttype is str:
            return f'{name} if isinstance({name}, str) else check.str_param({name}, "{name}")'
        elif ttype is float:
            return f'{name} if isinstance({name}, float) else check.float_param({name}, "{name}")'
        elif ttype is int:
            return f'{name} if isinstance({name}, int) else check.int_param({name}, "{name}")'
        elif ttype is bool:
            return f'{name} if isinstance({name}, bool) else check.bool_param({name}, "{name}")'
        elif ttype is Any:
            return name  # no-op

        # fallback to inst
        inst_type = _coerce_type(ttype, eval_ctx)
        if inst_type:
            it = _name(inst_type)
            return (
                f'{name} if isinstance({name}, {it}) else check.inst_param({name}, "{name}", {it})'
            )
        else:
            return name  # no-op
    elif origin is Literal:
        return f'check.literal_param({name}, "{name}", {args})'
    elif origin is Callable or origin is collections.abc.Callable:
        return f'check.callable_param({name}, "{name}")'
    else:
        if _is_annotated(origin, args):
            _process_annotated(ttype, args, eval_ctx)
            return build_check_call_str(args[0], f"{name}", eval_ctx)

        pair_left, pair_right = _container_pair_args(args, eval_ctx)
        single = _container_single_arg(args, eval_ctx)

        # containers
        if origin is list:
            return f'check.list_param({name}, "{name}", {_name(single)})'
        elif origin is dict:
            return f'check.dict_param({name}, "{name}", {_name(pair_left)}, {_name(pair_right)})'
        elif origin is set:
            return f'check.set_param({name}, "{name}", {_name(single)})'
        elif origin is collections.abc.Sequence:
            return f'check.sequence_param({name}, "{name}", {_name(single)})'
        elif origin is collections.abc.Iterable:
            return f'check.iterable_param({name}, "{name}", {_name(single)})'
        elif origin is collections.abc.Mapping:
            return f'check.mapping_param({name}, "{name}", {_name(pair_left)}, {_name(pair_right)})'
        elif origin is collections.abc.Set:
            return f'check.set_param({name}, "{name}", {_name(single)})'
        elif origin in (UnionType, Union):
            # optional
            if pair_right is type(None):
                inner_origin = get_origin(pair_left)
                # optional scalar
                if inner_origin is None:
                    if pair_left is str:
                        return f'{name} if {name} is None or isinstance({name}, str) else check.opt_str_param({name}, "{name}")'
                    elif pair_left is float:
                        return f'{name} if {name} is None or isinstance({name}, float) else check.opt_float_param({name}, "{name}")'
                    elif pair_left is int:
                        return f'{name} if {name} is None or isinstance({name}, int) else check.opt_int_param({name}, "{name}")'
                    elif pair_left is bool:
                        return f'{name} if {name} is None or isinstance({name}, bool) else check.opt_bool_param({name}, "{name}")'

                    # fallback to opt_inst
                    inst_type = _coerce_type(pair_left, eval_ctx)
                    it = _name(inst_type)
                    if inst_type:
                        return f'{name} if {name} is None or isinstance({name}, {it}) else check.opt_inst_param({name}, "{name}", {it})'
                    else:
                        return name  # no-op

                # optional container
                else:
                    inner_args = get_args(pair_left)
                    inner_pair_left, inner_pair_right = _container_pair_args(inner_args, eval_ctx)
                    inner_single = _container_single_arg(inner_args, eval_ctx)
                    if inner_origin is list:
                        return f'{name} if {name} is None else check.opt_nullable_list_param({name}, "{name}", {_name(inner_single)})'
                    elif inner_origin is dict:
                        return f'{name} if {name} is None else check.opt_nullable_dict_param({name}, "{name}", {_name(inner_pair_left)}, {_name(inner_pair_right)})'
                    elif inner_origin is set:
                        return f'{name} if {name} is None else check.opt_nullable_set_param({name}, "{name}", {_name(inner_single)})'
                    elif inner_origin is collections.abc.Sequence:
                        return f'{name} if {name} is None else check.opt_nullable_sequence_param({name}, "{name}", {_name(inner_single)})'
                    elif inner_origin is collections.abc.Iterable:
                        return f'{name} if {name} is None else check.opt_nullable_iterable_param({name}, "{name}", {_name(inner_single)})'
                    elif inner_origin is collections.abc.Mapping:
                        return f'{name} if {name} is None else check.opt_nullable_mapping_param({name}, "{name}", {_name(inner_pair_left)}, {_name(inner_pair_right)})'
                    elif inner_origin is collections.abc.Set:
                        return f'{name} if {name} is None else check.opt_nullable_set_param({name}, "{name}", {_name(inner_single)})'
                    elif is_record(inner_origin):
                        it = _name(inner_origin)
                        return f'{name} if {name} is None or isinstance({name}, {it}) else check.opt_inst_param({name}, "{name}", {it})'
                    elif inner_origin is Callable or inner_origin is collections.abc.Callable:
                        return f'{name} if {name} is None else check.opt_callable_param({name}, "{name}")'

            # union
            else:
                tuple_types = _coerce_type(ttype, eval_ctx)
                if tuple_types is not None:
                    tt_name = _name(tuple_types)
                    return f'{name} if isinstance({name}, {tt_name}) else check.inst_param({name}, "{name}", {tt_name})'

        # origin is some other type, assume ttype is Generic representation
        else:
            inst_type = _coerce_type(ttype, eval_ctx)
            if inst_type:
                it = _name(inst_type)
                return f'{name} if isinstance({name}, {it}) else check.inst_param({name}, "{name}", {it})'

        failed(f"Unhandled {ttype}")


def build_args_and_assignment_strs(
    fn_args: Iterable[str],
    defaults: Mapping[str, Any],
    kw_only: bool,
) -> tuple[str, str]:
    """Utility function to create the arguments to the function as well as any
    assignment calls that need to happen for default values.
    """
    args = []
    set_calls = []
    for arg in fn_args:
        if arg in defaults:
            default = defaults[arg]
            if default is None:
                args.append(f"{arg} = None")
            # dont share class instance of default empty containers
            elif default == []:
                args.append(f"{arg} = None")
                set_calls.append(f"{arg} = {arg} if {arg} is not None else []")
            elif default == {}:
                args.append(f"{arg} = None")
                set_calls.append(f"{arg} = {arg} if {arg} is not None else {'{}'}")
            # fallback to direct reference if unknown
            else:
                args.append(f"{arg} = {INJECTED_DEFAULT_VALS_LOCAL_VAR}['{arg}']")
        else:
            args.append(arg)

    args_str = ""
    if args:
        args_str = f", {'*,' if kw_only else ''} {', '.join(args)}"

    set_calls_str = ""
    if set_calls:
        set_calls_str = "\n    ".join(set_calls)

    return args_str, set_calls_str



import logging
import os
import tempfile
from collections.abc import Iterable, Iterator, Mapping
from contextlib import contextmanager
from typing import Any, Callable, Optional, TypeVar

import click
import tomli

import dagster._check as check
from dagster._core.instance import DagsterInstance, InstanceRef
from dagster._core.instance.config import is_dagster_home_set
from dagster._core.remote_representation.external import RemoteRepository
from dagster._core.secrets.env_file import get_env_var_dict
from dagster._utils.env import environ

T_Callable = TypeVar("T_Callable", bound=Callable[..., Any])


def has_pyproject_dagster_block(path: str) -> bool:
    if not os.path.exists(path):
        return False
    with open(path, "rb") as f:
        data = tomli.load(f)
        if not isinstance(data, dict):
            return False

        return "dagster" in data.get("tool", {})


@contextmanager
def _inject_local_env_file(logger: logging.Logger) -> Iterator[None]:
    dotenv_file_path = os.getcwd()

    env_var_dict = get_env_var_dict(dotenv_file_path)

    with environ(env_var_dict):
        if len(env_var_dict):
            logger.info(
                "Loaded environment variables from .env file: "
                + ",".join([env_var for env_var in env_var_dict]),
            )
        yield


TMP_DAGSTER_HOME_PREFIX = ".tmp_dagster_home_"


@contextmanager
def _get_temporary_instance(cli_command: str, logger: logging.Logger) -> Iterator[DagsterInstance]:
    # make the temp dir in the cwd since default temp dir roots
    # have issues with FS notify-based event log watching
    # use a fixed prefix so that consumer projects can reliably manage,
    # e.g., add to .gitignore
    with tempfile.TemporaryDirectory(prefix=TMP_DAGSTER_HOME_PREFIX, dir=os.getcwd()) as tempdir:
        logger.info(
            f"Using temporary directory {tempdir} for storage. This will be removed when"
            f" {cli_command} exits."
        )
        logger.info(
            "To persist information across sessions, set the environment variable DAGSTER_HOME"
            " to a directory to use."
        )

        with DagsterInstance.from_ref(
            InstanceRef.from_dir(tempdir, config_dir=os.getcwd())
        ) as instance:
            yield instance


@contextmanager
def get_temporary_instance_for_cli(
    cli_command: str = "this command",
    logger: logging.Logger = logging.getLogger("dagster"),
) -> Iterator[DagsterInstance]:
    with _inject_local_env_file(logger):
        with _get_temporary_instance(cli_command, logger) as instance:
            yield instance


@contextmanager
def get_possibly_temporary_instance_for_cli(
    cli_command: str = "this command",
    instance_ref: Optional[InstanceRef] = None,
    logger: logging.Logger = logging.getLogger("dagster"),
) -> Iterator[DagsterInstance]:
    """Create an instance at the entrypoint for a dagster CLI command. Handles loading
    any environment variables from a local .env file and finding the right path to load
    the instance from. If DAGSTER_HOME is not set and an instance_ref is not passed in, will
    create a temporary dagster.yaml file under the cwd.

    Parameters:
    cli_command: The name of the command - optional, used for logging messages.
    instance_ref: For commands that delegate to other commands, they can pass in
        a serialized instance_ref object rather than loading it from a dagster.yaml file.
    logger: Customize any log messages emitted while creating the instance.
    """
    with _inject_local_env_file(logger):
        if instance_ref:
            with DagsterInstance.from_ref(instance_ref) as instance:
                yield instance
        elif is_dagster_home_set():
            with DagsterInstance.get() as instance:
                yield instance
        else:
            with _get_temporary_instance(cli_command, logger) as instance:
                yield instance


@contextmanager
def get_instance_for_cli(
    instance_ref: Optional[InstanceRef] = None,
    logger: logging.Logger = logging.getLogger("dagster"),
) -> Iterator[DagsterInstance]:
    """Create an instance at the entrypoint for a dagster CLI command. Handles loading
    any environment variables from a local .env file and finding the right path to load
    the instance from. Requires DAGSTER_HOME to be set or instance_ref to be passed in.

    Parameters:
    cli_command: The name of the command - optional, used for logging messages.
    instance_ref: For commands that delegate to other commands, they can pass in
        a serialized instance_ref object rather than loading it from a dagster.yaml file.
    logger: Customize any log messages emitted while creating the instance.
    """
    with _inject_local_env_file(logger):
        if instance_ref:
            with DagsterInstance.from_ref(instance_ref) as instance:
                yield instance
        else:
            with DagsterInstance.get() as instance:
                yield instance


def assert_no_remaining_opts(opts: Mapping[str, object]) -> None:
    if opts:
        check.failed(
            f"Unexpected options remaining: {list(opts.keys())}. Ensure that all options are extracted."
        )


def serialize_sorted_quoted(strings: Iterable[str]) -> str:
    return "[" + ", ".join([f"'{s}'" for s in sorted(list(strings))]) + "]"


def validate_dagster_home_is_set() -> None:
    if not is_dagster_home_set():
        raise click.UsageError(
            "The environment variable $DAGSTER_HOME is not set. Dagster requires this "
            "environment variable to be set to an existing directory in your filesystem "
            "that contains your dagster instance configuration file (dagster.yaml).\n"
            "You can resolve this error by exporting the environment variable."
            "For example, you can run the following command in your shell or "
            "include it in your shell configuration file:\n"
            '\texport DAGSTER_HOME="~/dagster_home"'
            "\n\n"
        )


def validate_repo_has_defined_sensors(repo: RemoteRepository) -> None:
    if not repo.get_sensors():
        raise click.UsageError(f"There are no sensors defined for repository {repo.name}.")


def validate_repo_has_defined_schedules(repo: RemoteRepository) -> None:
    if not repo.get_schedules():
        raise click.UsageError(f"There are no schedules defined for repository {repo.name}.")
